"""Mocked requests library."""

import json
from typing import Callable

import requests

from libdyson.cloud.account import DYSON_API_HEADERS, DYSON_API_HOST


class MockedRequests:
    """Mocked requests library."""

    def __init__(self):
        """Initialize the mock."""
        self.host = DYSON_API_HOST
        self._handlers = {}

    @property
    def country(self) -> str:
        """Return the country."""
        return self._country

    def register_handler(self, method: str, path: str, handler: Callable) -> None:
        """Register request handler."""
        self._handlers[(method, path)] = handler

    def request(
        self, method: str, url: str, headers=None, verify=True, **kwargs
    ) -> requests.Response:
        """Run mocked request function."""
        assert headers == DYSON_API_HEADERS
        assert url.startswith(self.host)
        path = url[len(self.host) :]
        response = requests.Response()
        if not (method, path) in self._handlers:
            response.status_code = 404
            return response
        status_code, payload = self._handlers[(method, path)](**kwargs)
        response.status_code = status_code
        if isinstance(payload, bytes):
            response._content = payload
        elif payload is not None:
            response.encoding = "utf-8"
            content = json.dumps(payload).encode("utf-8")
            response._content = content
        return response



"""Dyson device info."""

from typing import Optional

import attr

from .utils import decrypt_password


@attr.s(auto_attribs=True, frozen=True)
class DysonDeviceInfo:
    """Dyson device info."""

    active: Optional[bool]
    serial: str
    name: str
    version: str
    credential: str
    auto_update: bool
    new_version_available: bool
    product_type: str

    @classmethod
    def from_raw(cls, raw: dict):
        """Parse raw data."""
        return cls(
            raw["Active"] if "Active" in raw else None,
            raw["Serial"],
            raw["Name"],
            raw["Version"],
            decrypt_password(raw["LocalCredentials"]),
            raw["AutoUpdate"],
            raw["NewVersionAvailable"],
            raw["ProductType"],
        )



"""Utilities for cloud tests."""

import base64
import json

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from libdyson.cloud.utils import DYSON_ENCRYPTION_INIT_VECTOR, DYSON_ENCRYPTION_KEY


def encrypt_credential(serial: str, credential: str) -> str:
    """Encrypt device local credential."""
    data = {
        "serial": serial,
        "apPasswordHash": credential,
    }
    data = json.dumps(data)
    padded_length = (len(data) + 15) // 16 * 16
    if padded_length > len(data):
        n_pad = padded_length - len(data)
        data += chr(n_pad) * n_pad
    data = data.encode("utf-8")
    cipher = Cipher(
        algorithms.AES(DYSON_ENCRYPTION_KEY),
        modes.CBC(DYSON_ENCRYPTION_INIT_VECTOR),
    )
    encryptor = cipher.encryptor()
    encrypted = encryptor.update(data) + encryptor.finalize()
    return base64.b64encode(encrypted).decode("utf-8")



"""Dyson Pure Cool Link fan."""

from .const import AirQualityTarget
from .dyson_device import DysonFanDevice


class DysonPureCoolLink(DysonFanDevice):
    """Dyson Pure Cool Link device."""

    @property
    def fan_mode(self) -> str:
        """Return the fan mode of the fan."""
        return self._get_field_value(self._status, "fmod")

    @property
    def is_on(self) -> bool:
        """Return if the device is on."""
        return self.fan_mode in ["FAN", "AUTO"]

    @property
    def auto_mode(self) -> bool:
        """Return auto mode status."""
        return self.fan_mode == "AUTO"

    @property
    def oscillation(self) -> bool:
        """Return oscillation status."""
        return self._get_field_value(self._status, "oson") == "ON"

    @property
    def air_quality_target(self) -> AirQualityTarget:
        """Return air quality target."""
        return AirQualityTarget(self._get_field_value(self._status, "qtar"))

    @property
    def filter_life(self) -> int:
        """Return filter life in hours."""
        return int(self._get_field_value(self._status, "filf"))

    @property
    def particulates(self) -> int:
        """Return particulate matter in unknown unit."""
        return self._get_environmental_field_value("pact")

    @property
    def volatile_organic_compounds(self) -> int:
        """Return VOCs in unknown unit."""
        return self._get_environmental_field_value("vact")

    def turn_on(self) -> None:
        """Turn on the device."""
        self._set_configuration(fmod="FAN")

    def turn_off(self) -> None:
        """Turn off the device."""
        self._set_configuration(fmod="OFF")

    def _set_speed(self, speed: int) -> None:
        self._set_configuration(fmod="FAN", fnsp=f"{speed:04d}")

    def enable_auto_mode(self) -> None:
        """Turn on auto mode."""
        self._set_configuration(fmod="AUTO")

    def disable_auto_mode(self) -> None:
        """Turn off auto mode."""
        self._set_configuration(fmod="FAN")

    def enable_oscillation(self) -> None:
        """Turn on oscillation."""
        self._set_configuration(oson="ON")

    def disable_oscillation(self) -> None:
        """Turn off oscillation."""
        self._set_configuration(oson="OFF")

    def enable_continuous_monitoring(self) -> None:
        """Turn on continuous monitoring."""
        self._set_configuration(
            fmod=self.fan_mode,  # Seems fmod is required to make this work
            rhtm="ON",
        )

    def disable_continuous_monitoring(self) -> None:
        """Turn off continuous monitoring."""
        self._set_configuration(
            fmod=self.fan_mode,
            rhtm="OFF",
        )

    def set_air_quality_target(self, air_quality_target: AirQualityTarget) -> None:
        """Set air quality target."""
        self._set_configuration(qtar=air_quality_target.value)



"""Tests for Dyson 360 Eye vacuum."""
from unittest.mock import patch

import pytest

from libdyson import DEVICE_TYPE_360_EYE, Dyson360Eye, VacuumEyePowerMode

from . import CREDENTIAL, HOST, SERIAL
from .mocked_mqtt import MockedMQTT

STATUS = {
    "currentVacuumPowerMode": "fullPower",
    "defaultVacuumPowerMode": "fullPower",
}


@pytest.fixture(autouse=True)
def mqtt_client() -> MockedMQTT:
    """Return mocked mqtt client."""
    mocked_mqtt = MockedMQTT(
        HOST,
        SERIAL,
        CREDENTIAL,
        f"{DEVICE_TYPE_360_EYE}/{SERIAL}/command",
        f"{DEVICE_TYPE_360_EYE}/{SERIAL}/status",
        STATUS,
    )
    with patch("libdyson.dyson_device.mqtt.Client", mocked_mqtt.refersh):
        yield mocked_mqtt


def test_properties(mqtt_client: MockedMQTT):
    """Test properties of 360 Eye."""
    device = Dyson360Eye(SERIAL, CREDENTIAL)
    device.connect(HOST)
    assert device.power_mode == VacuumEyePowerMode.MAX

    new_status = {
        "currentVacuumPowerMode": "halfPower",
        "defaultVacuumPowerMode": "halfPower",
    }
    mqtt_client.state_change(new_status)
    assert device.power_mode == VacuumEyePowerMode.QUIET


@pytest.mark.parametrize(
    "command,command_args,msg,msg_data",
    [
        ("start", [], "START", {"fullCleanType": "immediate"}),
        ("pause", [], "PAUSE", {}),
        ("resume", [], "RESUME", {}),
        ("abort", [], "ABORT", {}),
        (
            "set_power_mode",
            [VacuumEyePowerMode.MAX],
            "STATE-SET",
            {"data": {"defaultVacuumPowerMode": "fullPower"}},
        ),
        (
            "set_power_mode",
            [VacuumEyePowerMode.QUIET],
            "STATE-SET",
            {"data": {"defaultVacuumPowerMode": "halfPower"}},
        ),
    ],
)
def test_command(
    mqtt_client: MockedMQTT, command: str, command_args: list, msg: str, msg_data: dict
):
    """Test commands of 360 Eye."""
    device = Dyson360Eye(SERIAL, CREDENTIAL)
    device.connect(HOST)

    func = getattr(device, command)
    func(*command_args)
    len(mqtt_client.commands) == 1
    payload = mqtt_client.commands[0]
    assert payload.pop("msg") == msg
    payload.pop("time")
    assert payload == msg_data



"""Test Dyson Python library cloud submodule."""

AUTH_ACCOUNT = "fe011372-8949-4fe4-a815-06a9992126e0"
AUTH_PASSWORD = "jvDII/gIG8MXCaWydBalT1Uo4KxM4NOfUaXqd5mwe6OKb5cjQJUyjC1+g0bVi2nJO4TN0ZS2PawMqc6+ITB4fA=="
AUTH_INFO = {
    "Account": AUTH_ACCOUNT,
    "Password": AUTH_PASSWORD,
}



"""Tests for Pure Cool."""

import pytest

from libdyson import DEVICE_TYPE_PURE_COOL
from libdyson.const import ENVIRONMENTAL_FAIL, ENVIRONMENTAL_INIT, ENVIRONMENTAL_OFF
from libdyson.dyson_pure_cool import DysonPureCool

from . import CREDENTIAL, HOST, SERIAL
from .mocked_mqtt import MockedMQTT
from .test_fan_device import assert_command

DEVICE_TYPE = DEVICE_TYPE_PURE_COOL

STATUS = {
    "mode-reason": "RAPP",
    "state-reason": "MODE",
    "dial": "OFF",
    "rssi": "-46",
    "channel": "1",
    "product-state": {
        "fpwr": "OFF",
        "fdir": "OFF",
        "auto": "OFF",
        "oscs": "OFF",
        "oson": "OIOF",
        "nmod": "OFF",
        "rhtm": "OFF",
        "fnst": "OFF",
        "ercd": "NONE",
        "wacd": "NONE",
        "nmdv": "0004",
        "fnsp": "AUTO",
        "bril": "0002",
        "corf": "ON",
        "cflr": "0100",
        "hflr": "0100",
        "sltm": "OFF",
        "osal": "0063",
        "osau": "0243",
        "ancp": "CUST",
    },
    "scheduler": {"srsc": "000000005b1792f0", "dstv": "0001", "tzid": "0001"},
}

ENVIRONMENTAL_DATA = {
    "data": {
        "tact": "OFF",
        "hact": "OFF",
        "pm25": "OFF",
        "pm10": "OFF",
        "va10": "INIT",
        "noxl": "FAIL",
        "p25r": "OFF",
        "p10r": "OFF",
        "sltm": "OFF",
    }
}


def test_properties(mqtt_client: MockedMQTT):
    """Test properties of Pure Cool Link."""
    device = DysonPureCool(SERIAL, CREDENTIAL, DEVICE_TYPE)
    device.connect(HOST)

    # Status
    assert device.is_on is False
    assert device.auto_mode is False
    assert device.oscillation is False
    assert device.oscillation_status is False
    assert device.oscillation_angle_low == 63
    assert device.oscillation_angle_high == 243
    assert device.front_airflow is False
    assert device.night_mode_speed == 4
    assert device.carbon_filter_life == 100
    assert device.hepa_filter_life == 100

    # Environmental
    assert device.particulate_matter_2_5 == ENVIRONMENTAL_OFF
    assert device.particulate_matter_10 == ENVIRONMENTAL_OFF
    assert device.volatile_organic_compounds == ENVIRONMENTAL_INIT
    assert device.nitrogen_dioxide == ENVIRONMENTAL_FAIL

    new_status = {
        "mode-reason": "LAPP",
        "state-reason": "MODE",
        "product-state": {
            "fpwr": ["OFF", "ON"],
            "fdir": ["OFF", "ON"],
            "auto": ["OFF", "ON"],
            "oscs": ["OFF", "ON"],
            "oson": ["OIOF", "OION"],
            "nmod": ["OFF", "ON"],
            "rhtm": ["OFF", "ON"],
            "fnst": ["OFF", "ON"],
            "ercd": ["NONE", "NONE"],
            "wacd": ["NONE", "NONE"],
            "nmdv": ["0004", "0010"],
            "fnsp": ["AUTO", "AUTO"],
            "bril": ["0002", "0002"],
            "corf": ["ON", "ON"],
            "cflr": ["0100", "INV"],
            "hflr": ["0100", "80"],
            "sltm": ["OFF", "OFF"],
            "osal": ["0063", "0030"],
            "osau": ["0243", "0200"],
            "ancp": ["CUST", "CUST"],
        },
        "scheduler": {"srsc": "000000005b1792f0", "dstv": "0001", "tzid": "0001"},
    }
    mqtt_client.state_change(new_status)
    assert device.is_on is True
    assert device.auto_mode is True
    assert device.oscillation is True
    assert device.oscillation_status is True
    assert device.oscillation_angle_low == 30
    assert device.oscillation_angle_high == 200
    assert device.front_airflow is True
    assert device.night_mode_speed == 10
    assert device.carbon_filter_life is None
    assert device.hepa_filter_life == 80

    new_status = {
        "product-state": {
            "oson": ["ON", "OFF"],
        },
    }
    mqtt_client.state_change(new_status)
    assert device.oscillation is False
    new_status = {
        "product-state": {
            "oson": ["OFF", "ON"],
        },
    }
    mqtt_client.state_change(new_status)
    assert device.oscillation is True

    mqtt_client._environmental_data = {
        "data": {
            "tact": "2977",
            "hact": "0058",
            "pm25": "0009",
            "pm10": "0005",
            "va10": "0004",
            "noxl": "0011",
            "p25r": "0010",
            "p10r": "0009",
            "sltm": "OFF",
        }
    }
    device.request_environmental_data()
    assert device.particulate_matter_2_5 == 9
    assert device.particulate_matter_10 == 5
    assert device.volatile_organic_compounds == 4
    assert device.nitrogen_dioxide == 11


@pytest.mark.parametrize(
    "command,command_args,msg_data",
    [
        ("turn_on", [], {"fpwr": "ON"}),
        ("turn_off", [], {"fpwr": "OFF"}),
        ("set_speed", [3], {"fpwr": "ON", "fnsp": "0003"}),
        ("enable_auto_mode", [], {"auto": "ON"}),
        ("disable_auto_mode", [], {"auto": "OFF"}),
        (
            "enable_oscillation",
            [],
            {
                "oson": "OION",
                "fpwr": "ON",
                "ancp": "CUST",
                "osal": "0063",
                "osau": "0243",
            },
        ),
        (
            "enable_oscillation",
            [15, 15],
            {
                "oson": "OION",
                "fpwr": "ON",
                "ancp": "CUST",
                "osal": "0015",
                "osau": "0015",
            },
        ),
        (
            "enable_oscillation",
            [15, 45],
            {
                "oson": "OION",
                "fpwr": "ON",
                "ancp": "CUST",
                "osal": "0015",
                "osau": "0045",
            },
        ),
        ("disable_oscillation", [], {"oson": "OIOF"}),
        ("enable_continuous_monitoring", [], {"fpwr": "OFF", "rhtm": "ON"}),
        ("disable_continuous_monitoring", [], {"fpwr": "OFF", "rhtm": "OFF"}),
        ("enable_front_airflow", [], {"fdir": "ON"}),
        ("disable_front_airflow", [], {"fdir": "OFF"}),
    ],
)
def test_command(
    mqtt_client: MockedMQTT,
    command: str,
    command_args: list,
    msg_data: dict,
):
    """Test commands of Pure Cool Link."""
    assert_command(
        DysonPureCool(SERIAL, CREDENTIAL, DEVICE_TYPE),
        mqtt_client,
        command,
        command_args,
        msg_data,
    )


def test_oscillation_ON_OFF(mqtt_client: MockedMQTT):
    """Test enable and disable oscillation with ON/OFF instead of OION/OIOFF."""
    new_status = {
        "product-state": {
            "oson": ["OIOF", "ON"],
            "osal": ["0063", "0063"],
            "osau": ["0243", "0243"],
        },
    }
    assert_command(
        DysonPureCool(SERIAL, CREDENTIAL, DEVICE_TYPE),
        mqtt_client,
        "enable_oscillation",
        [],
        {
            "oson": "ON",
            "fpwr": "ON",
            "ancp": "CUST",
            "osal": "0063",
            "osau": "0243",
        },
        new_status,
    )
    assert_command(
        DysonPureCool(SERIAL, CREDENTIAL, DEVICE_TYPE),
        mqtt_client,
        "disable_oscillation",
        [],
        {"oson": "OFF"},
        new_status,
    )


@pytest.mark.parametrize(
    "angle_low,angle_high",
    [
        (3, 300),
        (5, 400),
        (300, 5),
        (5, 34),
    ],
)
def test_oscillation_invalid_data(
    mqtt_client: MockedMQTT, angle_low: int, angle_high: int
):
    """Test commands with invalid data."""
    device = DysonPureCool(SERIAL, CREDENTIAL, DEVICE_TYPE)
    device.connect(HOST)
    with pytest.raises(ValueError):
        device.enable_oscillation(angle_low, angle_high)
    assert len(mqtt_client.commands) == 0



"""Dyson Pure Humidify+Cool device."""

from typing import Optional

from .const import HumidifyOscillationMode, WaterHardness
from .dyson_pure_cool import DysonPureCoolBase

WATER_HARDNESS_ENUM_TO_STR = {
    WaterHardness.SOFT: "2025",
    WaterHardness.MEDIUM: "1350",
    WaterHardness.HARD: "0675",
}
WATER_HARDNESS_STR_TO_ENUM = {
    str_: enum for enum, str_ in WATER_HARDNESS_ENUM_TO_STR.items()
}


class DysonPureHumidifyCool(DysonPureCoolBase):
    """Dyson Pure Humidify+Cool device."""

    @property
    def oscillation(self) -> bool:
        """Return oscillation status."""
        return self._get_field_value(self._status, "oson") == "ON"

    @property
    def oscillation_mode(self) -> HumidifyOscillationMode:
        """Return oscillation mode."""
        return HumidifyOscillationMode(self._get_field_value(self._status, "ancp"))

    @property
    def humidification(self) -> bool:
        """Return if humidification is on."""
        return self._get_field_value(self._status, "hume") == "HUMD"

    @property
    def humidification_auto_mode(self) -> bool:
        """Return if humidification auto mode is on."""
        return self._get_field_value(self._status, "haut") == "ON"

    @property
    def target_humidity(self) -> int:
        """Return target humidity in percentage."""
        return int(self._get_field_value(self._status, "humt"))

    @property
    def auto_target_humidity(self) -> int:
        """Return humidification auto mode target humidity."""
        return int(self._get_field_value(self._status, "rect"))

    @property
    def water_hardness(self) -> WaterHardness:
        """Return the water hardness setting."""
        return WATER_HARDNESS_STR_TO_ENUM[self._get_field_value(self._status, "wath")]

    @property
    def time_until_next_clean(self) -> int:
        """Return the time remaining in hours before the next deep cleaning."""
        return int(self._get_field_value(self._status, "cltr"))

    @property
    def clean_time_remaining(self) -> int:
        """Return the time remaining in minutes before the cleaning finishes."""
        return int(self._get_field_value(self._status, "cdrr"))

    def enable_oscillation(
        self, oscillation_mode: Optional[HumidifyOscillationMode] = None
    ) -> None:
        """Turn on oscillation."""
        if oscillation_mode is None:
            oscillation_mode = self.oscillation_mode

        self._set_configuration(oson="ON", fpwr="ON", ancp=oscillation_mode.value)

    def disable_oscillation(self) -> None:
        """Turn off oscillation."""
        self._set_configuration(oson="OFF")

    def enable_humidification(self) -> None:
        """Enable humidification."""
        self._set_configuration(hume="HUMD")

    def disable_humidification(self) -> None:
        """Disable humidification."""
        self._set_configuration(hume="OFF")

    def enable_humidification_auto_mode(self) -> None:
        """Enable humidification auto mode."""
        self._set_configuration(haut="ON")

    def disable_humidification_auto_mode(self) -> None:
        """Disable humidification auto mode."""
        self._set_configuration(haut="OFF")

    def set_target_humidity(self, target_humidity: int) -> None:
        """Set target humidity."""
        self._set_configuration(humt=f"{target_humidity:04d}", haut="OFF")

    def set_water_hardness(self, water_hardness: WaterHardness) -> None:
        """Set water hardness."""
        self._set_configuration(wath=WATER_HARDNESS_ENUM_TO_STR[water_hardness])


class DysonPurifierHumidifyCoolFormaldehyde(DysonPureHumidifyCool):
    """Dyson Purifier Humidify+Cool Formaldehyde device."""

    @property
    def formaldehyde(self):
        """Return formaldehyde reading."""
        return int(self._get_environmental_field_value("hcho"))



"""Dyson test configuration."""

from unittest.mock import patch

import pytest

from . import CREDENTIAL, HOST, SERIAL
from .mocked_mqtt import MockedMQTT


@pytest.fixture()
def mqtt_client(request: pytest.FixtureRequest) -> MockedMQTT:
    """Return mocked mqtt client."""
    device_type = request.module.DEVICE_TYPE
    status = request.module.STATUS
    environmental_data = request.module.ENVIRONMENTAL_DATA
    mocked_mqtt = MockedMQTT(
        HOST,
        SERIAL,
        CREDENTIAL,
        f"{device_type}/{SERIAL}/command",
        f"{device_type}/{SERIAL}/status/current",
        status,
        environmental_data,
    )
    with patch("libdyson.dyson_device.mqtt.Client", mocked_mqtt.refersh), patch(
        "libdyson.dyson_device.TIMEOUT", 0
    ):
        yield mocked_mqtt



"""Dyson vacuum device."""

from typing import Optional, Tuple

from .const import CleaningType, VacuumState
from .dyson_device import DysonDevice


class DysonVacuumDevice(DysonDevice):
    """Dyson vacuum device."""

    @property
    def _status_topic(self) -> str:
        """MQTT status topic."""
        return f"{self.device_type}/{self._serial}/status"

    @property
    def state(self) -> VacuumState:
        """State of the device."""
        return VacuumState(
            self._status["state"]
            if "state" in self._status
            else self._status["newstate"]
        )

    @property
    def cleaning_type(self) -> Optional[CleaningType]:
        """Return the type of the current cleaning task."""
        cleaning_type = self._status["fullCleanType"]
        if cleaning_type == "":
            return None
        return CleaningType(cleaning_type)

    @property
    def cleaning_id(self) -> Optional[str]:
        """Return the id of the current cleaning task."""
        cleaning_id = self._status["cleanId"]
        if cleaning_id == "":
            return None
        return cleaning_id

    @property
    def battery_level(self) -> int:
        """Battery level of the device in percentage."""
        return self._status["batteryChargeLevel"]

    @property
    def position(self) -> Optional[Tuple[int, int]]:
        """Position (x, y) of the device."""
        if (
            "globalPosition" in self._status
            and len(self._status["globalPosition"]) == 2
        ):
            return tuple(self._status["globalPosition"])
        return None

    @property
    def is_charging(self) -> bool:
        """Whether the device is charging."""
        return self.state in [
            VacuumState.INACTIVE_CHARGING,
            VacuumState.INACTIVE_CHARGED,
            VacuumState.FULL_CLEAN_CHARGING,
            VacuumState.MAPPING_CHARGING,
        ]

    def _update_status(self, payload: dict) -> None:
        self._status = payload

    def pause(self) -> None:
        """Pause cleaning."""
        self._send_command("PAUSE")

    def resume(self) -> None:
        """Resume cleaning."""
        self._send_command("RESUME")

    def abort(self) -> None:
        """Abort cleaning."""
        self._send_command("ABORT")



"""Mocked paho-mqtt client."""

import json
from typing import Optional

import paho.mqtt.client as mqtt

from libdyson.utils import mqtt_time


class MockedMQTT:
    """Mocked paho-mqtt client."""

    def __init__(
        self,
        host: str,
        username: str,
        password: str,
        command_topic: str,
        status_topic: str,
        status: dict,
        environmental_data: Optional[dict] = None,
    ) -> None:
        """Initialize the client with expected values."""
        self._expected_host = host
        self._expected_username = username
        self._expected_password = password
        self._command_topic = command_topic
        self._status_topic = status_topic
        self._status = status
        self._environmental_data = environmental_data

    def refersh(self, protocol: str):
        """Refresh client state."""
        self._protocol = protocol
        self._username = None
        self._password = None

        self.on_connect = None
        self.on_disconnect = None
        self.on_message = None

        self._subscribed = False
        self.connected = False
        self.commands = []
        self.loop_started = False

        return self

    def username_pw_set(self, username: str, password: str) -> None:
        """Set username and password."""
        self._username = username
        self._password = password

    def connect_async(self, host: str) -> None:
        """Connect to the server asynchronously."""
        if host != self._expected_host:
            return
        if (
            self._username == self._expected_username
            and self._password == self._expected_password
        ):
            self.connected = True
            self.on_connect(self, None, None, 0)
        else:
            self.on_connect(self, None, None, 4)
            self.on_disconnect(self, None, 5)

    def disconnect(self) -> None:
        """Disconnect from the server."""
        self.connected = False
        self.on_disconnect(self, None, 0)

    def loop_start(self) -> None:
        """Start loop."""
        self.loop_started = True

    def loop_stop(self) -> None:
        """Stop loop."""
        self.loop_started = False

    def subscribe(self, topic: str) -> None:
        """Subscribe to a topic."""
        assert topic == self._status_topic
        self._subscribed = True

    def publish(self, topic: str, payload: str, qos: int = 0) -> None:
        """Publish to a topic."""
        assert topic == self._command_topic
        assert self._subscribed
        payload = json.loads(payload)
        message_type = payload["msg"]
        if message_type == "REQUEST-CURRENT-STATE":
            data = {"msg": "CURRENT-STATE"}
            data.update(self._status)
        elif message_type == "REQUEST-PRODUCT-ENVIRONMENT-CURRENT-SENSOR-DATA":
            data = {"msg": "ENVIRONMENTAL-CURRENT-SENSOR-DATA"}
            data.update(self._environmental_data)
        else:
            self.commands.append(payload)
            return

        message = mqtt.MQTTMessage(topic=self._status_topic.encode("utf-8"))
        data["time"] = mqtt_time()
        message.payload = json.dumps(data).encode("utf-8")
        self.on_message(self, None, message)

    def state_change(self, new_status) -> None:
        """Trigger a state change."""
        data = {
            "msg": "STATE-CHANGE",
            "time": mqtt_time(),
        }
        data.update(new_status)
        message = mqtt.MQTTMessage(topic=self._status_topic.encode("utf-8"))
        message.payload = json.dumps(data).encode("utf-8")
        self.on_message(self, None, message)



"""Test Dyson Python library."""
from typing import Type

import pytest

from libdyson import (
    DEVICE_TYPE_360_EYE,
    DEVICE_TYPE_360_HEURIST,
    DEVICE_TYPE_PURE_COOL,
    DEVICE_TYPE_PURE_COOL_FORMALDEHYDE,
    DEVICE_TYPE_PURE_COOL_DESK,
    DEVICE_TYPE_PURE_COOL_LINK,
    DEVICE_TYPE_PURE_COOL_LINK_DESK,
    DEVICE_TYPE_PURE_HOT_COOL,
    DEVICE_TYPE_PURE_HOT_COOL_NEW,
    DEVICE_TYPE_PURE_HOT_COOL_LINK,
    DEVICE_TYPE_PURE_HUMIDIFY_COOL,
    DEVICE_TYPE_PURIFIER_HUMIDIFY_COOL,
    DEVICE_TYPE_PURIFIER_HUMIDIFY_COOL_FORMALDEHYDE,
    DEVICE_TYPE_PURIFIER_HOT_COOL,
    Dyson360Eye,
    Dyson360Heurist,
    DysonDevice,
    DysonPureCool,
    DysonPureCoolFormaldehyde,
    DysonPureCoolLink,
    DysonPureHotCool,
    DysonPureHotCoolLink,
    DysonPureHumidifyCool,
    DysonPurifierHumidifyCoolFormaldehyde,
    get_device,
)

from . import CREDENTIAL, SERIAL


@pytest.mark.parametrize(
    "device_type,class_type",
    [
        (DEVICE_TYPE_360_EYE, Dyson360Eye),
        (DEVICE_TYPE_360_HEURIST, Dyson360Heurist),
        (DEVICE_TYPE_PURE_COOL_LINK_DESK, DysonPureCoolLink),
        (DEVICE_TYPE_PURE_COOL_LINK, DysonPureCoolLink),
        (DEVICE_TYPE_PURE_COOL, DysonPureCool),
        (DEVICE_TYPE_PURE_COOL_FORMALDEHYDE, DysonPureCoolFormaldehyde),
        (DEVICE_TYPE_PURE_COOL_DESK, DysonPureCool),
        (DEVICE_TYPE_PURE_HOT_COOL_LINK, DysonPureHotCoolLink),
        (DEVICE_TYPE_PURE_HOT_COOL, DysonPureHotCool),
        (DEVICE_TYPE_PURE_HOT_COOL_NEW, DysonPureHotCool),
        (DEVICE_TYPE_PURE_HUMIDIFY_COOL, DysonPureHumidifyCool),
        (DEVICE_TYPE_PURIFIER_HUMIDIFY_COOL, DysonPureHumidifyCool),
        (DEVICE_TYPE_PURIFIER_HUMIDIFY_COOL_FORMALDEHYDE, DysonPurifierHumidifyCoolFormaldehyde),
        (DEVICE_TYPE_PURIFIER_HOT_COOL, DysonPureHotCool),
    ],
)
def test_get_device(device_type: str, class_type: Type[DysonDevice]):
    """Test get_device."""
    device = get_device(SERIAL, CREDENTIAL, device_type)
    assert isinstance(device, class_type)
    assert device.serial == SERIAL


def test_get_device_unknown():
    """Test get_device with unknown type."""
    device = get_device(SERIAL, CREDENTIAL, "unknown")
    assert device is None



"""Dyson Python library exceptions."""


class DysonException(Exception):
    """Base class for exceptions."""


class DysonNetworkError(DysonException):
    """Represents network error."""


class DysonServerError(DysonException):
    """Represents Dyson server error."""


class DysonInvalidAccountStatus(DysonException):
    """Represents invalid account status."""


class DysonLoginFailure(DysonException):
    """Represents failure during logging in."""


class DysonAPIProvisionFailure(DysonException):
    """Represents failure during logging in."""


class DysonOTPTooFrequently(DysonException):
    """Represents requesting OTP code too frequently."""


class DysonAuthRequired(DysonException):
    """Represents not logged into could."""


class DysonInvalidAuth(DysonException):
    """Represents invalid authentication."""


class DysonConnectTimeout(DysonException):
    """Represents mqtt connection timeout."""


class DysonNotConnected(DysonException):
    """Represents mqtt not connected."""


class DysonInvalidCredential(DysonException):
    """Represents invalid mqtt credential."""


class DysonConnectionRefused(DysonException):
    """Represents mqtt connection refused by the server."""


class DysonFailedToParseWifiInfo(DysonException):
    """Represents failed to parse WiFi information."""



"""Tests for DysonAccount."""

from typing import Optional, Tuple

import pytest
import requests
from requests.auth import AuthBase, HTTPBasicAuth

from libdyson.cloud import DysonAccount
from libdyson.cloud.account import (
    API_PATH_DEVICES,
    API_PATH_EMAIL_REQUEST,
    API_PATH_EMAIL_VERIFY,
    API_PATH_MOBILE_REQUEST,
    API_PATH_MOBILE_VERIFY,
    API_PATH_USER_STATUS,
    API_PATH_PROVISION_APP,
    DYSON_API_HOST_CN,
    DysonAccountCN,
    HTTPBearerAuth,
)
from libdyson.exceptions import (
    DysonAuthRequired,
    DysonInvalidAccountStatus,
    DysonInvalidAuth,
    DysonLoginFailure,
    DysonNetworkError,
    DysonOTPTooFrequently,
    DysonServerError,
    DysonAPIProvisionFailure,
)

from . import AUTH_ACCOUNT, AUTH_INFO, AUTH_PASSWORD
from .mocked_requests import MockedRequests
from .utils import encrypt_credential

EMAIL = "user@example.com"
PASSWORD = "password"
REGION = "GB"
MOBILE = "+8613588888888"
OTP = "000000"
CHALLENGE_ID = "2b289d7f-1e0d-41e2-a0cb-56115eab6855"

BEARER_TOKEN = "BEARER_TOKEN"
AUTH_INFO_BEARER = {
    "token": BEARER_TOKEN,
    "tokenType": "Bearer",
    "account": AUTH_ACCOUNT,
}

DEVICE1_SERIAL = "NK6-CN-HAA0000A"
DEVICE1_NAME = "Device1"
DEVICE1_VERSION = "10.01.01CN"
DEVICE1_PRODUCT_TYPE = "475"
DEVICE1_CREDENTIAL = "aoWJM1kpL79MN2dPMlL5ysQv/APG+HAv+x3HDk0yuT3gMfgA3mLuil4O3d+q6CcyU+D1Hoir38soKoZHshYFeQ=="

DEVICE2_SERIAL = "JH1-US-HBB1111A"
DEVICE2_NAME = "Device2"
DEVICE2_VERSION = "11.02.02"
DEVICE2_PRODUCT_TYPE = "N223"
DEVICE2_CREDENTIAL = "KVjUpJoKwK7E9FDe5LN5JbUqDEfEDh5PlcNC7GJH1Ib2gGEpXwKEFszORFS8+tL8CNlvZTRmsUhf+kS37B7qAg=="

DEVICES = [
    {
        "Active": True,
        "Serial": DEVICE1_SERIAL,
        "Name": DEVICE1_NAME,
        "Version": DEVICE1_VERSION,
        "LocalCredentials": encrypt_credential(
            DEVICE1_SERIAL,
            DEVICE1_CREDENTIAL,
        ),
        "AutoUpdate": True,
        "NewVersionAvailable": False,
        "ProductType": DEVICE1_PRODUCT_TYPE,
        "ConnectionType": "wss",
    },
    {
        "Serial": DEVICE2_SERIAL,
        "Name": DEVICE2_NAME,
        "Version": DEVICE2_VERSION,
        "LocalCredentials": encrypt_credential(
            DEVICE2_SERIAL,
            DEVICE2_CREDENTIAL,
        ),
        "AutoUpdate": False,
        "NewVersionAvailable": True,
        "ProductType": DEVICE2_PRODUCT_TYPE,
        "ConnectionType": "wss",
    },
    {
        "Serial": "YS4-EU-MCA0660A",
        "Name": "Device3",
        "Version": "13.80.22",
        "LocalCredentials": None,
        "AutoUpdate": True,
        "NewVersionAvailable": False,
        "ProductType": "552",
        "ConnectionType": "wss",
    },
]


def _app_provision_handler(
        params: dict, json: dict, auth: Optional[AuthBase], **kwargs
) -> Tuple[int, Optional[str]]:
    return 200, '"5.0.21061"'


def _app_provision_handler_error(
        params: dict, json: dict, auth: Optional[AuthBase], **kwargs
) -> Tuple[int, Optional[str]]:
    return 404, ''


@pytest.fixture(autouse=True)
def mocked_requests(mocked_requests: MockedRequests) -> MockedRequests:
    """Return mocked requests library."""

    def _user_status_handler(
        params: dict, json: dict, auth: Optional[AuthBase], **kwargs
    ) -> Tuple[int, Optional[dict]]:
        assert auth is None
        assert params == {"country": REGION}
        if json["email"] == EMAIL:
            return (200, {"accountStatus": "ACTIVE"})
        return (200, {"accountStatus": "UNREGISTERED"})

    def _email_request_handler(
        json: dict, auth: Optional[AuthBase], **kwargs
    ) -> Tuple[int, Optional[dict]]:
        assert auth is None
        assert json == {
            "email": EMAIL,
        }
        return (200, {"challengeId": CHALLENGE_ID})

    def _email_verify_handler(
        json: dict, auth: Optional[AuthBase], **kwargs
    ) -> Tuple[int, Optional[dict]]:
        assert auth is None
        assert json["email"] == EMAIL
        assert json["challengeId"] == CHALLENGE_ID
        if json["password"] == PASSWORD and json["otpCode"] == OTP:
            return (200, AUTH_INFO_BEARER)
        return (400, None)

    def _mobile_request_handler(
        json: dict, auth: Optional[AuthBase], **kwargs
    ) -> Tuple[int, Optional[dict]]:
        assert auth is None
        assert json == {
            "mobile": MOBILE,
        }
        return (200, {"challengeId": CHALLENGE_ID})

    def _mobile_verify_handler(
        json: dict, auth: Optional[AuthBase], **kwargs
    ) -> Tuple[int, Optional[dict]]:
        assert auth is None
        assert json["mobile"] == MOBILE
        assert json["challengeId"] == CHALLENGE_ID
        if json["otpCode"] == OTP:
            return (200, AUTH_INFO_BEARER)
        return (400, None)

    def _devices_handler(
        auth: Optional[AuthBase], **kwargs
    ) -> Tuple[int, Optional[dict]]:
        if (
            not isinstance(auth, HTTPBasicAuth)
            or auth.username != AUTH_ACCOUNT
            or auth.password != AUTH_PASSWORD
        ) and (not isinstance(auth, HTTPBearerAuth) or auth.token != BEARER_TOKEN):
            return (401, None)
        return (200, DEVICES)

    mocked_requests.register_handler("POST", API_PATH_USER_STATUS, _user_status_handler)
    mocked_requests.register_handler("GET", API_PATH_PROVISION_APP, _app_provision_handler)
    mocked_requests.register_handler(
        "POST", API_PATH_EMAIL_REQUEST, _email_request_handler
    )
    mocked_requests.register_handler(
        "POST", API_PATH_EMAIL_VERIFY, _email_verify_handler
    )
    mocked_requests.register_handler(
        "POST", API_PATH_MOBILE_REQUEST, _mobile_request_handler
    )
    mocked_requests.register_handler(
        "POST", API_PATH_MOBILE_VERIFY, _mobile_verify_handler
    )
    mocked_requests.register_handler("GET", API_PATH_DEVICES, _devices_handler)
    return mocked_requests


def test_account_provision_api(mocked_requests: MockedRequests):
    account = DysonAccount()

    assert account.provision_api() is None

    mocked_requests.register_handler("GET", API_PATH_PROVISION_APP, _app_provision_handler_error)

    with pytest.raises(DysonAPIProvisionFailure):
        account.provision_api()

    mocked_requests.register_handler("GET", API_PATH_PROVISION_APP, _app_provision_handler)

def test_account():
    """Test account functionalities."""
    account = DysonAccount()

    # Incorrect email
    with pytest.raises(DysonInvalidAccountStatus):
        account.login_email_otp("unregistered@example.com", REGION)
    assert account.auth_info is None
    with pytest.raises(DysonAuthRequired):
        account.devices()

    # Incorrect password
    with pytest.raises(DysonLoginFailure):
        verify = account.login_email_otp(EMAIL, REGION)
        verify(OTP, "wrong_pass")
    assert account.auth_info is None

    # Incorrect OTP
    with pytest.raises(DysonLoginFailure):
        verify = account.login_email_otp(EMAIL, REGION)
        verify("999999", PASSWORD)
    assert account.auth_info is None

    # Login succeed
    verify = account.login_email_otp(EMAIL, REGION)
    verify(OTP, PASSWORD)
    assert account.auth_info == AUTH_INFO_BEARER

    # Devices
    devices = account.devices()
    assert devices[0].active is True
    assert devices[0].serial == DEVICE1_SERIAL
    assert devices[0].name == DEVICE1_NAME
    assert devices[0].version == DEVICE1_VERSION
    assert devices[0].credential == DEVICE1_CREDENTIAL
    assert devices[0].product_type == DEVICE1_PRODUCT_TYPE
    assert devices[0].auto_update is True
    assert devices[0].new_version_available is False
    assert devices[1].active is None
    assert devices[1].serial == DEVICE2_SERIAL
    assert devices[1].name == DEVICE2_NAME
    assert devices[1].version == DEVICE2_VERSION
    assert devices[1].credential == DEVICE2_CREDENTIAL
    assert devices[1].product_type == DEVICE2_PRODUCT_TYPE
    assert devices[1].auto_update is False
    assert devices[1].new_version_available is True


def test_account_auth_info():
    """Test initialize account with auth info."""
    account = DysonAccount(AUTH_INFO)
    devices = account.devices()
    assert len(devices) == 2

    # Invalid auth
    account = DysonAccount(
        {
            "Account": "invalid",
            "Password": "invalid",
        },
    )
    with pytest.raises(DysonInvalidAuth):
        account.devices()

    # No auth
    account = DysonAccount()
    with pytest.raises(DysonAuthRequired):
        account.devices()


def test_login_email_request_too_frequently(mocked_requests: MockedRequests):
    """Test request for otp code too frequently."""

    def _handle_email_request(
        json: dict, auth: Optional[AuthBase], **kwargs
    ) -> Tuple[int, Optional[dict]]:
        assert auth is None
        return (429, None)

    mocked_requests.register_handler(
        "POST", API_PATH_EMAIL_REQUEST, _handle_email_request
    )

    account = DysonAccount()
    with pytest.raises(DysonOTPTooFrequently):
        account.login_email_otp(EMAIL, REGION)


def test_login_mobile(mocked_requests: MockedRequests):
    """Test logging into account using phone number and otp code."""
    mocked_requests.host = DYSON_API_HOST_CN

    account = DysonAccountCN()
    verify = account.login_mobile_otp(MOBILE)

    # Incorrect OTP
    with pytest.raises(DysonLoginFailure):
        verify("111111")
    assert account.auth_info is None

    # Correct OTP
    verify(OTP)
    assert account.auth_info == AUTH_INFO_BEARER
    account.devices()


def test_login_mobile_request_too_frequently(mocked_requests: MockedRequests):
    """Test request for otp code too frequently."""

    def _handle_mobile_request(
        json: dict, auth: Optional[AuthBase], **kwargs
    ) -> Tuple[int, Optional[dict]]:
        assert auth is None
        return (429, None)

    mocked_requests.host = DYSON_API_HOST_CN
    mocked_requests.register_handler(
        "POST", API_PATH_MOBILE_REQUEST, _handle_mobile_request
    )

    account = DysonAccountCN()
    with pytest.raises(DysonOTPTooFrequently):
        account.login_mobile_otp(MOBILE)


def test_account_auth_info_bearer(mocked_requests: MockedRequests):
    """Test initialize account with bearer auth info."""
    mocked_requests.host = DYSON_API_HOST_CN
    account = DysonAccountCN(AUTH_INFO_BEARER)
    devices = account.devices()
    assert len(devices) == 2

    # Old auth
    account = DysonAccountCN(AUTH_INFO)
    devices = account.devices()
    assert len(devices) == 2

    # Invalid auth
    account = DysonAccountCN(
        {
            "token": "invalid",
            "tokenType": "Bearer",
            "account": "invalid",
        },
    )
    with pytest.raises(DysonInvalidAuth):
        account.devices()

    # No auth
    account = DysonAccountCN()
    with pytest.raises(DysonAuthRequired):
        account.devices()

    # Not supported auth info
    account = DysonAccountCN({"token": "TOKEN", "tokenType": "Custom"})
    with pytest.raises(DysonAuthRequired):
        account.devices()


def test_network_error(mocked_requests: MockedRequests):
    """Test network error handling."""

    def _handler_network_error(**kwargs):
        raise requests.RequestException

    mocked_requests.register_handler(
        "POST", API_PATH_EMAIL_REQUEST, _handler_network_error
    )
    mocked_requests.register_handler("GET", API_PATH_DEVICES, _handler_network_error)

    account = DysonAccount()
    with pytest.raises(DysonNetworkError):
        account.login_email_otp(EMAIL, REGION)
    account = DysonAccount(AUTH_INFO)
    with pytest.raises(DysonNetworkError):
        account.devices()


def test_server_error(mocked_requests: MockedRequests):
    """Test cloud server error handling."""

    def _handler_network_error(**kwargs):
        return (500, None)

    mocked_requests.register_handler(
        "POST", API_PATH_EMAIL_REQUEST, _handler_network_error
    )
    mocked_requests.register_handler("GET", API_PATH_DEVICES, _handler_network_error)

    account = DysonAccount()
    with pytest.raises(DysonServerError):
        account.login_email_otp(EMAIL, REGION)
    account = DysonAccount(AUTH_INFO)
    with pytest.raises(DysonServerError):
        account.devices()



"""Tests for utils."""

from libdyson.const import DEVICE_TYPE_360_EYE
from libdyson.utils import get_mqtt_info_from_wifi_info


def test_get_mqtt_info_from_wifi_info():
    """Test the function to get mqtt info from wifi info."""
    assert get_mqtt_info_from_wifi_info("JH1-US-GDA0001A", "z2jks80tmz") == (
        "JH1-US-GDA0001A",
        "wcosm2mJlB57tHnsxp3i8CiSz8H13J4i4p6Jw2SjdW3c+u+AZtG3PNmZ/ldaFS+Auubl5QRC/z3Lk4D3j+DQNw==",
        DEVICE_TYPE_360_EYE,
    )
    assert get_mqtt_info_from_wifi_info("360EYE-JH1-US-GDA0001A", "z2jks80tmz") == (
        "JH1-US-GDA0001A",
        "wcosm2mJlB57tHnsxp3i8CiSz8H13J4i4p6Jw2SjdW3c+u+AZtG3PNmZ/ldaFS+Auubl5QRC/z3Lk4D3j+DQNw==",
        DEVICE_TYPE_360_EYE,
    )
    assert get_mqtt_info_from_wifi_info("DYSON-NN3-EU-HWL4729E-475", "hjkj3gjask") == (
        "NN3-EU-HWL4729E",
        "Tp+0uDTy8scsa/PnYOeBVljZJiqzIZG+A0zOe4fiRJWXVKNOI4mpnFX7gDeEl5MEO10KFDbvVn4/4hiE2vpwiw==",
        "475",
    )
    assert get_mqtt_info_from_wifi_info("DYSON-SZ1-AU-MMZ2666D-455A", "sgrjjnsk01") == (
        "SZ1-AU-MMZ2666D",
        "sJKBwMhAGU9nAfdiHh4cXXCAm2E+YYPXGU6xp1NFcxX86f7NcRzs75QwCreZKt1SF1dQ9JoCrgj1bSsduaDcZA==",
        "455",
    )



"""Tests for Dyson Pure Hot+Cool Link device."""

import pytest

from libdyson import DEVICE_TYPE_PURE_HOT_COOL_LINK, DysonPureHotCoolLink

from . import CREDENTIAL, HOST, SERIAL
from .mocked_mqtt import MockedMQTT
from .test_fan_device import assert_command
from .test_pure_cool_link import ENVIRONMENTAL_DATA  # noqa: F401

DEVICE_TYPE = DEVICE_TYPE_PURE_HOT_COOL_LINK

STATUS = {"product-state": {"ffoc": "ON"}}


def test_properties(mqtt_client: MockedMQTT):
    """Test properties of Pure Hot+Cool Link."""
    device = DysonPureHotCoolLink(SERIAL, CREDENTIAL, DEVICE_TYPE)
    device.connect(HOST)

    # Status
    assert device.focus_mode is True

    new_status = {"product-state": {"ffoc": ["ON", "OFF"]}}
    mqtt_client.state_change(new_status)
    assert device.focus_mode is False


@pytest.mark.parametrize(
    "command,command_args,msg_data",
    [
        ("enable_focus_mode", [], {"ffoc": "ON"}),
        ("disable_focus_mode", [], {"ffoc": "OFF"}),
    ],
)
def test_command(
    mqtt_client: MockedMQTT,
    command: str,
    command_args: list,
    msg_data: dict,
):
    """Test commands of Pure Hot+Cool Link."""
    assert_command(
        DysonPureHotCoolLink(SERIAL, CREDENTIAL, DEVICE_TYPE),
        mqtt_client,
        command,
        command_args,
        msg_data,
    )



"""Dyson cloud client."""

from .account import DysonAccount, DysonAccountCN  # noqa: F401
from .cloud_360_eye import DysonCloud360Eye  # noqa: F401
from .cloud_device import DysonCloudDevice  # noqa: F401
from .device_info import DysonDeviceInfo  # noqa: F401
from .regions import REGIONS  # noqa: F401



"""Dyson 360 Heurist vacuum robot."""

from typing import Optional

from .const import DEVICE_TYPE_360_HEURIST, CleaningMode, VacuumHeuristPowerMode
from .dyson_vacuum_device import DysonVacuumDevice


class Dyson360Heurist(DysonVacuumDevice):
    """Dyson 360 Heurist device."""

    @property
    def device_type(self) -> str:
        """Return the device type."""
        return DEVICE_TYPE_360_HEURIST

    @property
    def current_power_mode(self) -> VacuumHeuristPowerMode:
        """Return current power mode."""
        return VacuumHeuristPowerMode(self._status["currentVacuumPowerMode"])

    @property
    def default_power_mode(self) -> VacuumHeuristPowerMode:
        """Return default power mode."""
        return VacuumHeuristPowerMode(self._status["defaultVacuumPowerMode"])

    @property
    def current_cleaning_mode(self) -> CleaningMode:
        """Return current cleaning mode."""
        return CleaningMode(self._status["currentCleaningMode"])

    @property
    def default_cleaning_mode(self) -> CleaningMode:
        """Return default cleaning mode."""
        return CleaningMode(self._status["defaultCleaningMode"])

    @property
    def is_bin_full(self) -> bool:
        """Return if the bin is full."""
        airways = self._status.get("faults", {}).get("AIRWAYS")
        if airways is None:
            return False
        return (
            airways.get("active") is True and airways.get("description") == "1.0.-1"
        )  # Not sure what this means

    def _send_command(self, command: str, data: Optional[dict] = None):
        if data is None:
            data = {}
        data["mode-reason"] = "LAPP"
        super()._send_command(command, data)

    def start_all_zones(self) -> None:
        """Start cleaning of all zones."""
        self._send_command(
            "START", {"cleaningMode": "global", "fullCleanType": "immediate"}
        )

    def set_default_power_mode(self, power_mode: VacuumHeuristPowerMode) -> None:
        """Set default power mode."""
        self._send_command(
            "STATE-SET",
            {"defaults": {"defaultVacuumPowerMode": power_mode.value}},
        )



"""Dyson Pure Cool fan."""

from abc import abstractmethod
from typing import Optional

from .dyson_device import DysonFanDevice


class DysonPureCoolBase(DysonFanDevice):
    """Dyson Pure Cool series base class."""

    @property
    def is_on(self) -> bool:
        """Return if the device is on."""
        return self._get_field_value(self._status, "fpwr") == "ON"

    @property
    def auto_mode(self) -> bool:
        """Return auto mode status."""
        return self._get_field_value(self._status, "auto") == "ON"

    @property
    @abstractmethod
    def oscillation(self) -> bool:
        """Return oscillation status."""

    @property
    def oscillation_status(self) -> bool:
        """Return the status of oscillation."""
        return self._get_field_value(self._status, "oscs") == "ON"

    @property
    def front_airflow(self) -> bool:
        """Return if airflow from front is on."""
        return self._get_field_value(self._status, "fdir") == "ON"

    @property
    def night_mode_speed(self) -> int:
        """Return speed in night mode."""
        return int(self._get_field_value(self._status, "nmdv"))

    @property
    def carbon_filter_life(self) -> Optional[int]:
        """Return carbon filter life in percentage."""
        filter_life = self._get_field_value(self._status, "cflr")
        if filter_life == "INV":
            return None
        return int(filter_life)

    @property
    def hepa_filter_life(self) -> Optional[int]:
        """Return HEPA filter life in percentage."""
        return int(self._get_field_value(self._status, "hflr"))

    @property
    def particulate_matter_2_5(self):
        """Return PM 2.5 in micro grams per cubic meter."""
        return int(self._get_environmental_field_value("pm25"))

    @property
    def particulate_matter_10(self):
        """Return PM 2.5 in micro grams per cubic meter."""
        return int(self._get_environmental_field_value("pm10"))

    @property
    def volatile_organic_compounds(self):
        """Return VOCs in micro grams per cubic meter."""
        return int(self._get_environmental_field_value("va10"))

    @property
    def nitrogen_dioxide(self):
        """Return nitrogen dioxide level in micro grams per cubic meter."""
        return int(self._get_environmental_field_value("noxl"))

    def turn_on(self) -> None:
        """Turn on the device."""
        self._set_configuration(fpwr="ON")

    def turn_off(self) -> None:
        """Turn off the device."""
        self._set_configuration(fpwr="OFF")

    def _set_speed(self, speed: int) -> None:
        self._set_configuration(fpwr="ON", fnsp=f"{speed:04d}")

    def enable_auto_mode(self) -> None:
        """Turn on auto mode."""
        self._set_configuration(auto="ON")

    def disable_auto_mode(self) -> None:
        """Turn off auto mode."""
        self._set_configuration(auto="OFF")

    def enable_continuous_monitoring(self) -> None:
        """Turn on continuous monitoring."""
        self._set_configuration(
            fpwr="ON" if self.is_on else "OFF",  # Not sure about this
            rhtm="ON",
        )

    def disable_continuous_monitoring(self) -> None:
        """Turn off continuous monitoring."""
        self._set_configuration(
            fpwr="ON" if self.is_on else "OFF",
            rhtm="OFF",
        )

    def enable_front_airflow(self) -> None:
        """Turn on front airflow."""
        self._set_configuration(fdir="ON")

    def disable_front_airflow(self) -> None:
        """Turn off front airflow."""
        self._set_configuration(fdir="OFF")


class DysonPureCool(DysonPureCoolBase):
    """Dyson Pure Cool device."""

    @property
    def oscillation(self) -> bool:
        """Return oscillation status."""
        # Seems some devices use OION/OIOF while others uses ON/OFF
        # https://github.com/shenxn/ha-dyson/issues/22
        return self._get_field_value(self._status, "oson") in ["OION", "ON"]

    @property
    def oscillation_angle_low(self) -> int:
        """Return oscillation low angle."""
        return int(self._get_field_value(self._status, "osal"))

    @property
    def oscillation_angle_high(self) -> int:
        """Return oscillation high angle."""
        return int(self._get_field_value(self._status, "osau"))

    def enable_oscillation(
        self,
        angle_low: Optional[int] = None,
        angle_high: Optional[int] = None,
    ) -> None:
        """Turn on oscillation."""
        if angle_low is None:
            angle_low = self.oscillation_angle_low
        if angle_high is None:
            angle_high = self.oscillation_angle_high

        if not 5 <= angle_low <= 355:
            raise ValueError("angle_low must be between 5 and 355")
        if not 5 <= angle_high <= 355:
            raise ValueError("angle_high must be between 5 and 355")
        if angle_low != angle_high and angle_low + 30 > angle_high:
            raise ValueError(
                "angle_high must be either equal to angle_low or at least 30 larger than angle_low"
            )

        current_oscillation_raw = self._get_field_value(self._status, "oson")
        if current_oscillation_raw in ["OION", "OIOF"]:
            oson = "OION"
        else:
            oson = "ON"
        self._set_configuration(
            oson=oson,
            fpwr="ON",
            ancp="CUST",
            osal=f"{angle_low:04d}",
            osau=f"{angle_high:04d}",
        )

    def disable_oscillation(self) -> None:
        """Turn off oscillation."""
        current_oscillation_raw = self._get_field_value(self._status, "oson")
        if current_oscillation_raw in ["OION", "OIOF"]:
            oson = "OIOF"
        else:
            oson = "OFF"
        self._set_configuration(oson=oson)


class DysonPureCoolFormaldehyde(DysonPureCool):
    """This model is compatible with PureCool but has one additional sensor."""

    @property
    def formaldehyde(self) -> Optional[int]:
        """Return formaldehyde reading."""
        # Dyson documentation for Dyson Pure Cool Formaldehyde refers to
        # the formaldehyde reading as HCHO (pp5).
        # https://www.dyson.com/content/dam/dyson/maintenance/user-guides/en_US/airtreatment/purifiers/TP09/497043-01.pdf
        #
        # H-CHO is also a common way to refer to formaldehyde.
        #
        # This is part of environmental data as per:
        # https://github.com/seanrees/prometheus-dyson/issues/13#issue-923525150
        return self._get_environmental_field_value("hcho")


"""Dyson Python library."""

from typing import Optional

from .const import (
    DEVICE_TYPE_360_EYE,
    DEVICE_TYPE_360_HEURIST,
    DEVICE_TYPE_PURE_COOL,
    DEVICE_TYPE_PURIFIER_COOL,
    DEVICE_TYPE_PURE_COOL_FORMALDEHYDE,
    DEVICE_TYPE_PURE_COOL_DESK,
    DEVICE_TYPE_PURE_COOL_LINK,
    DEVICE_TYPE_PURE_COOL_LINK_DESK,
    DEVICE_TYPE_PURE_HOT_COOL,
    DEVICE_TYPE_PURE_HOT_COOL_NEW,
    DEVICE_TYPE_PURE_HOT_COOL_LINK,
    DEVICE_TYPE_PURE_HUMIDIFY_COOL,
    DEVICE_TYPE_PURIFIER_HUMIDIFY_COOL,
    DEVICE_TYPE_PURIFIER_HUMIDIFY_COOL_FORMALDEHYDE,
    DEVICE_TYPE_PURIFIER_HOT_COOL,
)
from .const import CleaningMode  # noqa: F401
from .const import CleaningType  # noqa: F401
from .const import DEVICE_TYPE_NAMES  # noqa: F401
from .const import HumidifyOscillationMode  # noqa: F401
from .const import MessageType  # noqa: F401
from .const import VacuumEyePowerMode  # noqa: F401
from .const import VacuumHeuristPowerMode  # noqa: F401
from .const import VacuumState  # noqa: F401
from .const import WaterHardness  # noqa: F401
from .discovery import DysonDiscovery  # noqa: F401
from .dyson_360_eye import Dyson360Eye
from .dyson_360_heurist import Dyson360Heurist
from .dyson_device import DysonDevice
from .dyson_pure_cool import DysonPureCool
from .dyson_pure_cool import DysonPureCoolFormaldehyde
from .dyson_pure_cool_link import DysonPureCoolLink
from .dyson_pure_hot_cool import DysonPureHotCool
from .dyson_pure_hot_cool_link import DysonPureHotCoolLink
from .dyson_pure_humidify_cool import DysonPureHumidifyCool, DysonPurifierHumidifyCoolFormaldehyde
from .utils import get_mqtt_info_from_wifi_info  # noqa: F401


def get_device(serial: str, credential: str, device_type: str) -> Optional[DysonDevice]:
    """Get a new DysonDevice instance."""
    if device_type == DEVICE_TYPE_360_EYE:
        return Dyson360Eye(serial, credential)
    if device_type == DEVICE_TYPE_360_HEURIST:
        return Dyson360Heurist(serial, credential)
    if device_type in [
        DEVICE_TYPE_PURE_COOL_LINK_DESK,
        DEVICE_TYPE_PURE_COOL_LINK,
    ]:
        return DysonPureCoolLink(serial, credential, device_type)
    if device_type in [
        DEVICE_TYPE_PURE_COOL,
        DEVICE_TYPE_PURIFIER_COOL,
        DEVICE_TYPE_PURE_COOL_DESK,
    ]:
        return DysonPureCool(serial, credential, device_type)
    if device_type == DEVICE_TYPE_PURE_COOL_FORMALDEHYDE:
        return DysonPureCoolFormaldehyde(serial, credential, device_type)
    if device_type == DEVICE_TYPE_PURE_HOT_COOL_LINK:
        return DysonPureHotCoolLink(serial, credential, device_type)
    if device_type in [
        DEVICE_TYPE_PURE_HOT_COOL,
        DEVICE_TYPE_PURE_HOT_COOL_NEW,
        DEVICE_TYPE_PURIFIER_HOT_COOL,
    ]:
        return DysonPureHotCool(serial, credential, device_type)
    if device_type in [
        DEVICE_TYPE_PURE_HUMIDIFY_COOL,
        DEVICE_TYPE_PURIFIER_HUMIDIFY_COOL
    ]:
        return DysonPureHumidifyCool(serial, credential, device_type)
    if device_type == DEVICE_TYPE_PURIFIER_HUMIDIFY_COOL_FORMALDEHYDE:
        return DysonPurifierHumidifyCoolFormaldehyde(serial, credential, device_type)
    return None



"""Constants for Dyson Python library."""
from enum import Enum, auto

DEVICE_TYPE_360_EYE = "N223"
DEVICE_TYPE_360_HEURIST = "276"
DEVICE_TYPE_PURE_COOL_LINK = "475"
DEVICE_TYPE_PURE_COOL_LINK_DESK = "469"
DEVICE_TYPE_PURE_COOL = "438"
DEVICE_TYPE_PURIFIER_COOL = "438K"
DEVICE_TYPE_PURE_COOL_FORMALDEHYDE = "438E"
DEVICE_TYPE_PURE_COOL_DESK = "520"
DEVICE_TYPE_PURE_HUMIDIFY_COOL = "358"
DEVICE_TYPE_PURIFIER_HUMIDIFY_COOL = "358"
DEVICE_TYPE_PURIFIER_HUMIDIFY_COOL_FORMALDEHYDE = "358E"
DEVICE_TYPE_PURE_HOT_COOL_LINK = "455"
DEVICE_TYPE_PURE_HOT_COOL = "527"
DEVICE_TYPE_PURE_HOT_COOL_NEW = "527E"
DEVICE_TYPE_PURIFIER_HOT_COOL = "527K"

DEVICE_TYPE_NAMES = {
    DEVICE_TYPE_360_EYE: "360 Eye robot vacuum",
    DEVICE_TYPE_360_HEURIST: "360 Heurist robot vacuum",
    DEVICE_TYPE_PURE_COOL: "Pure Cool",
    DEVICE_TYPE_PURIFIER_COOL: "Purifier Cool",
    DEVICE_TYPE_PURE_COOL_FORMALDEHYDE: "Pure Cool Formaldehyde",
    DEVICE_TYPE_PURE_COOL_DESK: "Pure Cool Desk",
    DEVICE_TYPE_PURE_COOL_LINK: "Pure Cool Link",
    DEVICE_TYPE_PURE_COOL_LINK_DESK: "Pure Cool Link Desk",
    DEVICE_TYPE_PURE_HOT_COOL: "Pure Hot+Cool",
    DEVICE_TYPE_PURE_HOT_COOL_NEW: "Pure Hot+Cool (New)",
    DEVICE_TYPE_PURE_HOT_COOL_LINK: "Pure Hot+Cool Link",
    DEVICE_TYPE_PURE_HUMIDIFY_COOL: "Pure Humidify+Cool",
    DEVICE_TYPE_PURIFIER_HUMIDIFY_COOL: "Purifier Humidify+Cool",
    DEVICE_TYPE_PURIFIER_HUMIDIFY_COOL_FORMALDEHYDE: "Purifier Humidify+Cool Formaldehyde",
    DEVICE_TYPE_PURIFIER_HOT_COOL: "Purifier Hot+Cool",
}

ENVIRONMENTAL_OFF = -1
ENVIRONMENTAL_INIT = -2
ENVIRONMENTAL_FAIL = -3


class MessageType(Enum):
    """Update message type."""

    STATE = auto()
    ENVIRONMENTAL = auto()


class AirQualityTarget(Enum):
    """Air Quality Target."""

    OFF = "OFF"
    GOOD = "0004"
    SENSITIVE = "0003"
    DEFAULT = "0002"
    VERY_SENSITIVE = "0001"


class HumidifyOscillationMode(Enum):
    """Pure Humidify+Cool oscillation mode."""

    DEGREE_45 = "0045"
    DEGREE_90 = "0090"
    BREEZE = "BRZE"
    CUST = "CUST"


class WaterHardness(Enum):
    """Water Hardness."""

    SOFT = "Soft"
    MEDIUM = "Medium"
    HARD = "Hard"


class VacuumState(Enum):
    """Dyson vacuum state."""

    FAULT_CALL_HELPLINE = "FAULT_CALL_HELPLINE"
    FAULT_CONTACT_HELPLINE = "FAULT_CONTACT_HELPLINE"
    FAULT_CRITICAL = "FAULT_CRITICAL"
    FAULT_GETTING_INFO = "FAULT_GETTING_INFO"
    FAULT_LOST = "FAULT_LOST"
    FAULT_ON_DOCK = "FAULT_ON_DOCK"
    FAULT_ON_DOCK_CHARGED = "FAULT_ON_DOCK_CHARGED"
    FAULT_ON_DOCK_CHARGING = "FAULT_ON_DOCK_CHARGING"
    FAULT_REPLACE_ON_DOCK = "FAULT_REPLACE_ON_DOCK"
    FAULT_RETURN_TO_DOCK = "FAULT_RETURN_TO_DOCK"
    FAULT_RUNNING_DIAGNOSTIC = "FAULT_RUNNING_DIAGNOSTIC"
    FAULT_USER_RECOVERABLE = "FAULT_USER_RECOVERABLE"
    FULL_CLEAN_ABANDONED = "FULL_CLEAN_ABANDONED"
    FULL_CLEAN_ABORTED = "FULL_CLEAN_ABORTED"
    FULL_CLEAN_CHARGING = "FULL_CLEAN_CHARGING"
    FULL_CLEAN_DISCOVERING = "FULL_CLEAN_DISCOVERING"
    FULL_CLEAN_FINISHED = "FULL_CLEAN_FINISHED"
    FULL_CLEAN_INITIATED = "FULL_CLEAN_INITIATED"
    FULL_CLEAN_NEEDS_CHARGE = "FULL_CLEAN_NEEDS_CHARGE"
    FULL_CLEAN_PAUSED = "FULL_CLEAN_PAUSED"
    FULL_CLEAN_RUNNING = "FULL_CLEAN_RUNNING"
    FULL_CLEAN_TRAVERSING = "FULL_CLEAN_TRAVERSING"
    INACTIVE_CHARGED = "INACTIVE_CHARGED"
    INACTIVE_CHARGING = "INACTIVE_CHARGING"
    INACTIVE_DISCHARGING = "INACTIVE_DISCHARGING"
    MAPPING_ABORTED = "MAPPING_ABORTED"
    MAPPING_CHARGING = "MAPPING_CHARGING"
    MAPPING_FINISHED = "MAPPING_FINISHED"
    MAPPING_INITIATED = "MAPPING_INITIATED"
    MAPPING_NEEDS_CHARGE = "MAPPING_NEEDS_CHARGE"
    MAPPING_PAUSED = "MAPPING_PAUSED"
    MAPPING_RUNNING = "MAPPING_RUNNING"


class VacuumEyePowerMode(Enum):
    """Dyson 360 Eye power mode."""

    QUIET = "halfPower"
    MAX = "fullPower"


class VacuumHeuristPowerMode(Enum):
    """Dyson 360 Heurist power mode."""

    QUIET = "1"
    HIGH = "2"
    MAX = "3"


class CleaningType(Enum):
    """Vacuum cleaning type."""

    IMMEDIATE = "immediate"
    MANUAL = "manual"
    SCHEDULED = "scheduled"


class CleaningMode(Enum):
    """Vacuum cleaning mode."""

    GLOBAL = "global"
    ZONE_CONFIGURED = "zoneConfigured"



"""Dyson cloud account client."""

import pathlib
from typing import Callable, List, Optional

import requests
from requests.auth import AuthBase, HTTPBasicAuth

from libdyson.exceptions import (
    DysonAuthRequired,
    DysonInvalidAccountStatus,
    DysonInvalidAuth,
    DysonLoginFailure,
    DysonNetworkError,
    DysonOTPTooFrequently,
    DysonServerError,
    DysonAPIProvisionFailure,
)

from .device_info import DysonDeviceInfo

DYSON_API_HOST = "https://appapi.cp.dyson.com"
DYSON_API_HOST_CN = "https://appapi.cp.dyson.cn"
DYSON_API_HEADERS = {
    "User-Agent": "android client"
}

API_PATH_PROVISION_APP = "/v1/provisioningservice/application/Android/version"
API_PATH_USER_STATUS = "/v3/userregistration/email/userstatus"
API_PATH_EMAIL_REQUEST = "/v3/userregistration/email/auth"
API_PATH_EMAIL_VERIFY = "/v3/userregistration/email/verify"
API_PATH_MOBILE_REQUEST = "/v3/userregistration/mobile/auth"
API_PATH_MOBILE_VERIFY = "/v3/userregistration/mobile/verify"
API_PATH_DEVICES = "/v2/provisioningservice/manifest"

FILE_PATH = pathlib.Path(__file__).parent.absolute()

class HTTPBearerAuth(AuthBase):
    """Attaches HTTP Bearder Authentication to the given Request object."""

    def __init__(self, token):
        """Initialize the auth."""
        self.token = token

    def __eq__(self, other):
        """Return if equal."""
        return self.token == getattr(other, "token", None)

    def __ne__(self, other):
        """Return if not equal."""
        return not self == other

    def __call__(self, r):
        """Attach the authentication."""
        r.headers["Authorization"] = f"Bearer {self.token}"
        return r


class DysonAccount:
    """Dyson account."""

    _HOST = DYSON_API_HOST

    def __init__(
        self,
        auth_info: Optional[dict] = None,
    ):
        """Create a new Dyson account."""
        self._auth_info = auth_info

    @property
    def auth_info(self) -> Optional[dict]:
        """Return the authentication info."""
        return self._auth_info

    @property
    def _auth(self) -> Optional[AuthBase]:
        if self.auth_info is None:
            return None
        # Although basic auth is no longer used by new logins,
        # we still need this for backward capability to already
        # stored auth info.
        if "Password" in self.auth_info:
            return HTTPBasicAuth(
                self.auth_info["Account"],
                self.auth_info["Password"],
            )
        elif self.auth_info.get("tokenType") == "Bearer":
            return HTTPBearerAuth(self.auth_info["token"])
        return None

    def request(
        self,
        method: str,
        path: str,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        auth: bool = True,
    ) -> requests.Response:
        """Make API request."""
        if auth and self._auth is None:
            raise DysonAuthRequired
        try:
            response = requests.request(
                method,
                self._HOST + path,
                params=params,
                json=data,
                headers=DYSON_API_HEADERS,
                auth=self._auth if auth else None,
                verify=True,
            )
        except requests.RequestException:
            raise DysonNetworkError
        if response.status_code in [401, 403]:
            raise DysonInvalidAuth
        if 500 <= response.status_code < 600:
            raise DysonServerError
        return response

    def provision_api(self) -> None:
        """Provision the client connection to the API

        Calls an app provisioning API. This is expected by the Dyson App API and makes the API server ready to accept
        other API calls from the current IP Address.

        Basically, this unlocks the API - the return value is not needed, and we don't need to save any cookies or
        session tokens. It seems like the API Server sets some internal flag allowing API Calls from a specific address
        based solely on this endpoint being called.

        This isn't likely to be a security measure. It returns a version number in a json-encoded string: `"5.0.21061"`
        and that is likely consumed by an app. Presumably, an official Dyson mobile app could check the version against
        some internal expected value and, for example, prompt a user that it is outdated and direct them to the app
        store to download a new version in order to continue working.
        """

        response = self.request(
            "GET",
            API_PATH_PROVISION_APP,
            params=None,
            data=None,
            auth=False,
        )

        if response.status_code != 200:
            raise DysonAPIProvisionFailure

    def login_email_otp(self, email: str, region: str) -> Callable[[str, str], dict]:
        """Login using email and OTP code."""
        self.provision_api()

        # Check account status. This tells us whether an account is active or not.
        response = self.request(
            "POST",
            API_PATH_USER_STATUS,
            params={"country": region},
            data={"email": email},
            auth=False,
        )

        jsonRes = response.json()

        account_status = jsonRes["accountStatus"]
        if account_status != "ACTIVE":
            raise DysonInvalidAccountStatus(account_status)

        response = self.request(
            "POST",
            API_PATH_EMAIL_REQUEST,
            params={"country": region, "culture": "en-US"},
            data={"email": email},
            auth=False,
        )
        if response.status_code == 429:
            raise DysonOTPTooFrequently

        challenge_id = response.json()["challengeId"]

        def _verify(otp_code: str, password: str):
            response = self.request(
                "POST",
                API_PATH_EMAIL_VERIFY,
                data={
                    "email": email,
                    "password": password,
                    "challengeId": challenge_id,
                    "otpCode": otp_code,
                },
                auth=False,
            )
            if response.status_code == 400:
                raise DysonLoginFailure
            body = response.json()
            self._auth_info = body
            return self._auth_info

        return _verify

    def devices(self) -> List[DysonDeviceInfo]:
        self.provision_api()
        """Get device info from cloud account."""
        devices = []
        response = self.request("GET", API_PATH_DEVICES)
        for raw in response.json():
            if raw.get("LocalCredentials") is None:
                # Lightcycle lights don't have LocalCredentials.
                # They're not supported so just skip.
                # See https://github.com/shenxn/libdyson/issues/2 for more info
                continue
            devices.append(DysonDeviceInfo.from_raw(raw))
        return devices


class DysonAccountCN(DysonAccount):
    """Dyson account in Mainland China."""

    _HOST = DYSON_API_HOST_CN

    def login_mobile_otp(self, mobile: str) -> Callable[[str], dict]:
        self.provision_api()

        """Login using phone number and OTP code."""
        response = self.request(
            "POST",
            API_PATH_MOBILE_REQUEST,
            data={"mobile": mobile},
            auth=False,
        )
        if response.status_code == 429:
            raise DysonOTPTooFrequently

        challenge_id = response.json()["challengeId"]

        def _verify(otp_code: str):
            response = self.request(
                "POST",
                API_PATH_MOBILE_VERIFY,
                data={
                    "mobile": mobile,
                    "challengeId": challenge_id,
                    "otpCode": otp_code,
                },
                auth=False,
            )
            if response.status_code == 400:
                raise DysonLoginFailure
            body = response.json()
            self._auth_info = body
            return self._auth_info

        return _verify



"""Tests for Dyson Pure Humidity+Cool device."""

import pytest

from libdyson import (
    DEVICE_TYPE_PURE_HUMIDIFY_COOL,
    DysonPureHumidifyCool,
    HumidifyOscillationMode,
    WaterHardness,
)

from . import CREDENTIAL, HOST, SERIAL
from .mocked_mqtt import MockedMQTT
from .test_fan_device import assert_command
from .test_pure_cool_link import ENVIRONMENTAL_DATA  # noqa: F401

DEVICE_TYPE = DEVICE_TYPE_PURE_HUMIDIFY_COOL

STATUS = {
    "product-state": {
        "oson": "ON",
        "ancp": "BRZE",
        "hume": "HUMD",
        "haut": "ON",
        "humt": "0050",
        "rect": "0080",
        "wath": "2025",
        "cltr": "1853",
        "cdrr": "0060",
    }
}


def test_properties(mqtt_client: MockedMQTT):
    """Test properties of Pure Hot+Cool Link."""
    device = DysonPureHumidifyCool(SERIAL, CREDENTIAL, DEVICE_TYPE)
    device.connect(HOST)

    assert device.oscillation is True
    assert device.oscillation_mode == HumidifyOscillationMode.BREEZE
    assert device.humidification is True
    assert device.humidification_auto_mode is True
    assert device.target_humidity == 50
    assert device.auto_target_humidity == 80
    assert device.water_hardness == WaterHardness.SOFT
    assert device.time_until_next_clean == 1853
    assert device.clean_time_remaining == 60

    new_status = {
        "product-state": {
            "oson": ["ON", "OFF"],
            "ancp": ["BRZE", "0045"],
            "hume": ["HUMD", "OFF"],
            "haut": ["ON", "OFF"],
            "humt": ["0050", "0030"],
            "rect": ["0080", "0050"],
            "wath": ["2025", "0675"],
            "cltr": ["1853", "1800"],
            "cdrr": ["0060", "0050"],
        }
    }
    mqtt_client.state_change(new_status)
    assert device.oscillation is False
    assert device.oscillation_mode == HumidifyOscillationMode.DEGREE_45
    assert device.humidification is False
    assert device.humidification_auto_mode is False
    assert device.target_humidity == 30
    assert device.auto_target_humidity == 50
    assert device.water_hardness == WaterHardness.HARD
    assert device.time_until_next_clean == 1800
    assert device.clean_time_remaining == 50


@pytest.mark.parametrize(
    "command,command_args,msg_data",
    [
        ("enable_oscillation", [], {"oson": "ON", "fpwr": "ON", "ancp": "BRZE"}),
        (
            "enable_oscillation",
            [HumidifyOscillationMode.DEGREE_45],
            {"oson": "ON", "fpwr": "ON", "ancp": "0045"},
        ),
        (
            "enable_oscillation",
            [HumidifyOscillationMode.DEGREE_90],
            {"oson": "ON", "fpwr": "ON", "ancp": "0090"},
        ),
        (
            "enable_oscillation",
            [HumidifyOscillationMode.BREEZE],
            {"oson": "ON", "fpwr": "ON", "ancp": "BRZE"},
        ),
        ("disable_oscillation", [], {"oson": "OFF"}),
        ("enable_humidification", [], {"hume": "HUMD"}),
        ("disable_humidification", [], {"hume": "OFF"}),
        ("enable_humidification_auto_mode", [], {"haut": "ON"}),
        ("disable_humidification_auto_mode", [], {"haut": "OFF"}),
        ("set_target_humidity", [50], {"humt": "0050", "haut": "OFF"}),
        ("set_water_hardness", [WaterHardness.SOFT], {"wath": "2025"}),
        ("set_water_hardness", [WaterHardness.MEDIUM], {"wath": "1350"}),
        ("set_water_hardness", [WaterHardness.HARD], {"wath": "0675"}),
    ],
)
def test_command(
    mqtt_client: MockedMQTT,
    command: str,
    command_args: list,
    msg_data: dict,
):
    """Test commands of Pure Hot+Cool Link."""
    assert_command(
        DysonPureHumidifyCool(SERIAL, CREDENTIAL, DEVICE_TYPE),
        mqtt_client,
        command,
        command_args,
        msg_data,
    )



"""Tests for Dyson vacuum base class."""
from unittest.mock import patch

import pytest

from libdyson import DEVICE_TYPE_360_EYE, CleaningType, Dyson360Eye, VacuumState

from . import CREDENTIAL, HOST, SERIAL
from .mocked_mqtt import MockedMQTT

STATUS = {
    "state": "INACTIVE_CHARGED",
    "fullCleanType": "",
    "cleanId": "",
    "globalPosition": [0, 0],
    "batteryChargeLevel": 100,
}


@pytest.fixture(autouse=True)
def mqtt_client() -> MockedMQTT:
    """Return mocked mqtt client."""
    mocked_mqtt = MockedMQTT(
        HOST,
        SERIAL,
        CREDENTIAL,
        f"{DEVICE_TYPE_360_EYE}/{SERIAL}/command",
        f"{DEVICE_TYPE_360_EYE}/{SERIAL}/status",
        STATUS,
    )
    with patch("libdyson.dyson_device.mqtt.Client", mocked_mqtt.refersh):
        yield mocked_mqtt


def test_properties(mqtt_client: MockedMQTT):
    """Test properties of vaccums."""
    device = Dyson360Eye(SERIAL, CREDENTIAL)
    device.connect(HOST)

    assert device.state == VacuumState.INACTIVE_CHARGED
    assert device.cleaning_type is None
    assert device.cleaning_id is None
    assert device.position == (0, 0)
    assert device.is_charging is True
    assert device.battery_level == 100

    clean_id = "b599d00f-6f3b-401a-9c05-69877251e843"
    new_status = {
        "oldstate": "INACTIVE_CHARGED",
        "newstate": "FULL_CLEAN_RUNNING",
        "fullCleanType": "immediate",
        "cleanId": clean_id,
        "globalPosition": [],
        "batteryChargeLevel": 30,
    }
    mqtt_client.state_change(new_status)
    assert device.state == VacuumState.FULL_CLEAN_RUNNING
    assert device.cleaning_type == CleaningType.IMMEDIATE
    assert device.cleaning_id == clean_id
    assert device.position is None
    assert device.is_charging is False
    assert device.battery_level == 30



"""Utility functions for Dyson Python library."""

import base64
import hashlib
import re
import time
from typing import Tuple

from .const import DEVICE_TYPE_360_EYE
from .exceptions import DysonFailedToParseWifiInfo

# For some devices, the model in WiFi SSID is not the same as the model for MQTT.
# The model on Dyson Cloud always matches the one used for MQTT.
_DEVICE_TYPE_MAP = {
    "455A": "455",
}


def mqtt_time():
    """Return current time string for mqtt messages."""
    return time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())


def get_credential_from_wifi_password(wifi_password: str) -> str:
    """Calculate MQTT credential from WiFi password."""
    hash_ = hashlib.sha512()
    hash_.update(wifi_password.encode("utf-8"))
    return base64.b64encode(hash_.digest()).decode("utf-8")


def get_mqtt_info_from_wifi_info(
    wifi_ssid: str, wifi_password: str
) -> Tuple[str, str, str]:
    """Get MQTT information from WiFi information."""
    result = re.match(r"^(360EYE-)?(?P<serial>[0-9A-Z]{3}-[A-Z]{2}-[0-9A-Z]{8,})$", wifi_ssid)
    if result is not None:
        serial = result.group("serial")
        device_type = DEVICE_TYPE_360_EYE
    else:
        result = re.match(
            r"^DYSON-([0-9A-Z]{3}-[A-Z]{2}-[0-9A-Z]{8,})-([0-9]{3}[A-Z]?)$", wifi_ssid
        )
        if result is not None:
            serial = result.group(1)
            device_type = result.group(2)
            device_type = _DEVICE_TYPE_MAP.get(device_type, device_type)
        else:
            raise DysonFailedToParseWifiInfo

    credential = get_credential_from_wifi_password(wifi_password)
    return serial, credential, device_type



"""Tests for 360 Eye cloud client."""

from datetime import datetime, timedelta
from typing import Optional, Tuple

from requests.auth import AuthBase

from libdyson.cloud import DysonAccount
from libdyson.cloud.cloud_360_eye import CleaningType, DysonCloud360Eye

from . import AUTH_INFO
from .mocked_requests import MockedRequests

SERIAL = "JH1-US-HBB1111A"


def test_get_cleaning_history(mocked_requests: MockedRequests):
    """Test get cleaning history from the cloud."""
    cleaning1_id = "edcda2c9-5088-455e-b2ee-9422ef70afb2"
    cleaning2_id = "98cf2de1-190f-4e68-97b5-c57e7e0604d0"
    clean_history = {
        "TriviaMessage": "Your robot has cleaned 1000yds²",
        "TriviaArea": 800.1243,
        "Entries": [
            {
                "Clean": cleaning1_id,
                "Started": "2021-02-10T17:02:00",
                "Finished": "2021-02-10T17:02:10",
                "Area": 0.00,
                "Charges": 0,
                "Type": "Immediate",
                "IsInterim": False,
            },
            {
                "Clean": cleaning2_id,
                "Started": "2021-02-09T12:10:07",
                "Finished": "2021-02-09T14:14:11",
                "Area": 34.70,
                "Charges": 1,
                "Type": "Scheduled",
                "IsInterim": True,
            },
        ],
    }

    def _clean_history_handler(
        auth: Optional[AuthBase], **kwargs
    ) -> Tuple[int, Optional[dict]]:
        assert auth is not None
        return (0, clean_history)

    mocked_requests.register_handler(
        "GET", f"/v1/assets/devices/{SERIAL}/cleanhistory", _clean_history_handler
    )

    account = DysonAccount(AUTH_INFO)
    device = DysonCloud360Eye(account, SERIAL)
    cleaning_tasks = device.get_cleaning_history()
    assert len(cleaning_tasks) == 2
    task = cleaning_tasks[0]
    assert task.cleaning_id == cleaning1_id
    assert task.start_time == datetime(2021, 2, 10, 17, 2, 0)
    assert task.finish_time == datetime(2021, 2, 10, 17, 2, 10)
    assert task.cleaning_time == timedelta(seconds=10)
    assert task.area == 0.0
    assert task.charges == 0
    assert task.cleaning_type == CleaningType.Immediate
    assert task.is_interim is False
    task = cleaning_tasks[1]
    assert task.cleaning_id == cleaning2_id
    assert task.start_time == datetime(2021, 2, 9, 12, 10, 7)
    assert task.finish_time == datetime(2021, 2, 9, 14, 14, 11)
    assert task.cleaning_time == timedelta(hours=2, minutes=4, seconds=4)
    assert task.area == 34.7
    assert task.charges == 1
    assert task.cleaning_type == CleaningType.Scheduled
    assert task.is_interim is True


def test_get_cleaning_map(mocked_requests: MockedRequests):
    """Test get cleaning map from the cloud."""
    cleaning_id = "edcda2c9-5088-455e-b2ee-9422ef70afb2"
    cleaning_map = b"mocked_png_image"

    def _clean_history_handler(auth: Optional[AuthBase], **kwargs) -> Tuple[int, bytes]:
        assert auth is not None
        return (0, cleaning_map)

    mocked_requests.register_handler(
        "GET",
        f"/v1/mapvisualizer/devices/{SERIAL}/map/{cleaning_id}",
        _clean_history_handler,
    )

    account = DysonAccount(AUTH_INFO)
    device = DysonCloud360Eye(account, SERIAL)
    assert device.get_cleaning_map(cleaning_id) == cleaning_map

    # Non existed map
    assert device.get_cleaning_map("another_id") is None



"""Dyson device cloud client."""


from . import DysonAccount


class DysonCloudDevice:
    """Dyson device cloud client."""

    def __init__(self, account: DysonAccount, serial: str):
        """Initialize the client."""
        self._account = account
        self._serial = serial



"""Tests for Dyson heating device."""

import pytest

from libdyson import DEVICE_TYPE_PURE_HOT_COOL_LINK, DysonPureHotCoolLink

from . import CREDENTIAL, HOST, SERIAL
from .mocked_mqtt import MockedMQTT
from .test_fan_device import assert_command
from .test_pure_cool_link import ENVIRONMENTAL_DATA  # noqa: F401

DEVICE_TYPE = DEVICE_TYPE_PURE_HOT_COOL_LINK

STATUS = {
    "product-state": {
        "tilt": "OK",
        "hmax": "2805",
        "hmod": "HEAT",
        "hsta": "HEAT",
    }
}


def test_properties(mqtt_client: MockedMQTT):
    """Test properties of heating device."""
    device = DysonPureHotCoolLink(SERIAL, CREDENTIAL, DEVICE_TYPE)
    device.connect(HOST)

    # Status
    assert device.tilt is False
    assert device.heat_target == 280.5
    assert device.heat_mode_is_on is True
    assert device.heat_status_is_on is True

    new_status = {
        "product-state": {
            "tilt": ["OK", "TILT"],
            "hmax": ["2805", "2755"],
            "hmod": ["HEAT", "OFF"],
            "hsta": ["HEAT", "OFF"],
        }
    }
    mqtt_client.state_change(new_status)
    assert device.tilt is True
    assert device.heat_target == 275.5
    assert device.heat_mode_is_on is False
    assert device.heat_status_is_on is False


@pytest.mark.parametrize(
    "command,command_args,msg_data",
    [
        ("set_heat_target", [279.5], {"hmod": "HEAT", "hmax": "2795"}),
        ("enable_heat_mode", [], {"hmod": "HEAT"}),
        ("disable_heat_mode", [], {"hmod": "OFF"}),
    ],
)
def test_command(
    mqtt_client: MockedMQTT,
    command: str,
    command_args: list,
    msg_data: dict,
):
    """Test commands of heating device."""
    assert_command(
        DysonPureHotCoolLink(SERIAL, CREDENTIAL, DEVICE_TYPE),
        mqtt_client,
        command,
        command_args,
        msg_data,
    )


@pytest.mark.parametrize("heat_target", [273, 310.5])
def test_set_heat_target_invalid_value(mqtt_client: MockedMQTT, heat_target: float):
    """Test set heat target with invalid value."""
    device = DysonPureHotCoolLink(SERIAL, CREDENTIAL, DEVICE_TYPE)
    device.connect(HOST)
    with pytest.raises(ValueError):
        device.set_heat_target(heat_target)
    assert len(mqtt_client.commands) == 0



"""Tests for HTTPBearerAuth."""

from unittest.mock import MagicMock

import requests

from libdyson.cloud.account import HTTPBearerAuth


def test_bearer_auth():
    """Test HTTPBearerAuth."""
    auth1 = HTTPBearerAuth("token")
    auth2 = HTTPBearerAuth("token")
    auth3 = HTTPBearerAuth("token3")

    assert auth1 == auth2
    assert auth1 != auth3
    request = MagicMock(spec=requests.Request)
    request.headers = {"Key": "value"}
    request = auth1(request)
    assert request.headers == {
        "Key": "value",
        "Authorization": "Bearer token",
    }



"""Dyson Pure Hot+Cool Link device."""

from .dyson_device import DysonHeatingDevice
from .dyson_pure_cool_link import DysonPureCoolLink


class DysonPureHotCoolLink(DysonPureCoolLink, DysonHeatingDevice):
    """Dyson Pure Hot+Cool Link device."""

    @property
    def tilt(self) -> bool:
        """Return tilt status."""
        return self._get_field_value(self._status, "tilt") == "TILT"

    def enable_focus_mode(self) -> None:
        """Enable fan focus mode."""
        self._set_configuration(ffoc="ON")

    def disable_focus_mode(self) -> None:
        """Disable fan focus mode."""
        self._set_configuration(ffoc="OFF")



"""Dyson 360 Eye vacuum robot."""

from .const import DEVICE_TYPE_360_EYE, VacuumEyePowerMode
from .dyson_vacuum_device import DysonVacuumDevice


class Dyson360Eye(DysonVacuumDevice):
    """Dyson 360 Eye device."""

    @property
    def device_type(self) -> str:
        """Return the device type."""
        return DEVICE_TYPE_360_EYE

    @property
    def power_mode(self) -> VacuumEyePowerMode:
        """Power mode of the device."""
        return VacuumEyePowerMode(self._status["currentVacuumPowerMode"])

    def start(self) -> None:
        """Start cleaning."""
        self._send_command("START", {"fullCleanType": "immediate"})

    def set_power_mode(self, power_mode: VacuumEyePowerMode) -> None:
        """Set power mode."""
        self._send_command(
            "STATE-SET",
            {"data": {"defaultVacuumPowerMode": power_mode.value}},
        )



"""Dyson device."""
from abc import abstractmethod
import json
import logging
import threading
from typing import Any, Optional

import paho.mqtt.client as mqtt

from .const import (
    ENVIRONMENTAL_FAIL,
    ENVIRONMENTAL_INIT,
    ENVIRONMENTAL_OFF,
    MessageType,
)
from .exceptions import (
    DysonConnectionRefused,
    DysonConnectTimeout,
    DysonInvalidCredential,
    DysonNotConnected,
)
from .utils import mqtt_time

_LOGGER = logging.getLogger(__name__)

TIMEOUT = 10


class DysonDevice:
    """Base class for dyson devices."""

    def __init__(self, serial: str, credential: str):
        """Initialize the device."""
        self._serial = serial
        self._credential = credential
        self._mqtt_client = None
        self._connected = threading.Event()
        self._disconnected = threading.Event()
        self._status = None
        self._status_data_available = threading.Event()
        self._callbacks = []

    @property
    def serial(self) -> str:
        """Return the serial number of the device."""
        return self._serial

    @property
    def is_connected(self) -> bool:
        """Whether MQTT connection is active."""
        return self._connected.is_set()

    @property
    @abstractmethod
    def device_type(self) -> str:
        """Device type."""

    @property
    @abstractmethod
    def _status_topic(self) -> str:
        """MQTT status topic."""

    @property
    def _command_topic(self) -> str:
        """MQTT command topic."""
        return f"{self.device_type}/{self._serial}/command"

    def _request_first_data(self) -> bool:
        """Request and wait for first data."""
        self.request_current_status()
        return self._status_data_available.wait(timeout=TIMEOUT)

    def connect(self, host: str) -> None:
        """Connect to the device MQTT broker."""
        self._disconnected.clear()
        self._mqtt_client = mqtt.Client(protocol=mqtt.MQTTv31)
        self._mqtt_client.username_pw_set(self._serial, self._credential)
        error = None

        def _on_connect(client: mqtt.Client, userdata: Any, flags, rc):
            _LOGGER.debug("Connected with result code %d", rc)
            nonlocal error
            if rc == mqtt.CONNACK_REFUSED_BAD_USERNAME_PASSWORD:
                error = DysonInvalidCredential
            elif rc != mqtt.CONNACK_ACCEPTED:
                error = DysonConnectionRefused
            else:
                client.subscribe(self._status_topic)
            self._connected.set()

        def _on_disconnect(client, userdata, rc):
            _LOGGER.debug(f"Disconnected with result code {str(rc)}")

        self._disconnected.set()

        self._mqtt_client.on_connect = _on_connect
        self._mqtt_client.on_disconnect = _on_disconnect
        self._mqtt_client.on_message = self._on_message
        self._mqtt_client.connect_async(host)
        self._mqtt_client.loop_start()
        if self._connected.wait(timeout=TIMEOUT):
            if error is not None:
                self.disconnect()
                raise error

            _LOGGER.info("Connected to device %s", self._serial)
            if self._request_first_data():
                self._mqtt_client.on_connect = self._on_connect
                self._mqtt_client.on_disconnect = self._on_disconnect
                return

        # Close connection if timeout or connected but failed to get data
        self.disconnect()

        raise DysonConnectTimeout

    def disconnect(self) -> None:
        """Disconnect from the device."""
        self._connected.clear()
        self._mqtt_client.disconnect()
        if not self._disconnected.wait(timeout=TIMEOUT):
            _LOGGER.warning("Disconnect timed out")
        self._mqtt_client.loop_stop()
        self._mqtt_client = None

    def add_message_listener(self, callback) -> None:
        """Add a callback to receive update notification."""
        self._callbacks.append(callback)

    def remove_message_listener(self, callback) -> None:
        """Remove an existed callback."""
        if callback in self._callbacks:
            self._callbacks.remove(callback)

    def _on_connect(self, client: mqtt.Client, userdata: Any, flags, rc):
        _LOGGER.debug("Connected with result code %d", rc)
        self._disconnected.clear()
        self._connected.set()
        client.subscribe(self._status_topic)
        for callback in self._callbacks:
            callback(MessageType.STATE)

    def _on_disconnect(self, client, userdata, rc):
        _LOGGER.debug(f"Disconnected with result code {str(rc)}")
        self._connected.clear()
        self._disconnected.set()
        for callback in self._callbacks:
            callback(MessageType.STATE)

    def _on_message(self, client, userdata: Any, msg: mqtt.MQTTMessage):
        payload = json.loads(msg.payload.decode("utf-8"))
        self._handle_message(payload)

    def _handle_message(self, payload: dict) -> None:
        if payload["msg"] in ["CURRENT-STATE", "STATE-CHANGE"]:
            _LOGGER.debug("New state: %s", payload)
            self._update_status(payload)
            if not self._status_data_available.is_set():
                self._status_data_available.set()
            for callback in self._callbacks:
                callback(MessageType.STATE)

    @abstractmethod
    def _update_status(self, payload: dict) -> None:
        """Update the device status."""

    def _send_command(self, command: str, data: Optional[dict] = None):
        if not self.is_connected:
            raise DysonNotConnected
        if data is None:
            data = {}
        payload = {
            "msg": command,
            "time": mqtt_time(),
        }
        payload.update(data)
        self._mqtt_client.publish(self._command_topic, json.dumps(payload))

    def request_current_status(self):
        """Request current status."""
        if not self.is_connected:
            raise DysonNotConnected
        payload = {
            "msg": "REQUEST-CURRENT-STATE",
            "time": mqtt_time(),
        }
        self._mqtt_client.publish(self._command_topic, json.dumps(payload))


class DysonFanDevice(DysonDevice):
    """Dyson fan device."""

    def __init__(self, serial: str, credential: str, device_type: str):
        """Initialize the device."""
        super().__init__(serial, credential)
        self._device_type = device_type

        self._environmental_data = None
        self._environmental_data_available = threading.Event()

    @property
    def device_type(self) -> str:
        """Device type."""
        return self._device_type

    @property
    def _status_topic(self) -> str:
        """MQTT status topic."""
        return f"{self.device_type}/{self._serial}/status/current"

    @property
    def fan_state(self) -> bool:
        """Return if the fan is running."""
        return self._get_field_value(self._status, "fnst") == "FAN"

    @property
    def speed(self) -> Optional[int]:
        """Return fan speed."""
        speed = self._get_field_value(self._status, "fnsp")
        if speed == "AUTO":
            return None
        return int(speed)

    @property
    @abstractmethod
    def is_on(self) -> bool:
        """Return if the device is on."""

    @property
    @abstractmethod
    def auto_mode(self) -> bool:
        """Return auto mode status."""

    @property
    @abstractmethod
    def oscillation(self) -> bool:
        """Return oscillation status."""

    @property
    def night_mode(self) -> bool:
        """Return night mode status."""
        return self._get_field_value(self._status, "nmod") == "ON"

    @property
    def continuous_monitoring(self) -> bool:
        """Return standby monitoring status."""
        return self._get_field_value(self._status, "rhtm") == "ON"

    @property
    def error_code(self) -> str:
        """Return error code."""
        return self._get_field_value(self._status, "ercd")

    @property
    def warning_code(self) -> str:
        """Return warning code."""
        return self._get_field_value(self._status, "wacd")

    @property
    def humidity(self) -> int:
        """Return humidity in percentage."""
        return self._get_environmental_field_value("hact")

    @property
    def temperature(self) -> int:
        """Return temperature in kelvin."""
        return self._get_environmental_field_value("tact", divisor=10)

    @property
    @abstractmethod
    def volatile_organic_compounds(self) -> int:
        """Return VOCs."""

    @property
    def sleep_timer(self) -> int:
        """Return sleep timer in minutes."""
        return self._get_environmental_field_value("sltm")

    @staticmethod
    def _get_field_value(state, field):
        return state[field][1] if isinstance(state[field], list) else state[field]

    def _get_environmental_field_value(self, field, divisor=1):
        value = self._get_field_value(self._environmental_data, field)
        if value == "OFF":
            return ENVIRONMENTAL_OFF
        if value == "INIT":
            return ENVIRONMENTAL_INIT
        if value == "FAIL":
            return ENVIRONMENTAL_FAIL
        if divisor == 1:
            return int(value)
        return float(value) / divisor

    def _handle_message(self, payload: dict) -> None:
        super()._handle_message(payload)
        if payload["msg"] == "ENVIRONMENTAL-CURRENT-SENSOR-DATA":
            _LOGGER.debug("New environmental state: %s", payload)
            self._environmental_data = payload["data"]
            if not self._environmental_data_available.is_set():
                self._environmental_data_available.set()
            for callback in self._callbacks:
                callback(MessageType.ENVIRONMENTAL)

    def _update_status(self, payload: dict) -> None:
        self._status = payload["product-state"]

    def _set_configuration(self, **kwargs: dict) -> None:
        if not self.is_connected:
            raise DysonNotConnected
        payload = json.dumps(
            {
                "msg": "STATE-SET",
                "time": mqtt_time(),
                "mode-reason": "LAPP",
                "data": kwargs,
            }
        )
        self._mqtt_client.publish(self._command_topic, payload, 1)

    def _request_first_data(self) -> bool:
        """Request and wait for first data."""
        self.request_current_status()
        self.request_environmental_data()
        status_available = self._status_data_available.wait(timeout=TIMEOUT)
        environmental_available = self._environmental_data_available.wait(
            timeout=TIMEOUT
        )
        return status_available and environmental_available

    def request_environmental_data(self):
        """Request environmental sensor data."""
        if not self.is_connected:
            raise DysonNotConnected
        payload = {
            "msg": "REQUEST-PRODUCT-ENVIRONMENT-CURRENT-SENSOR-DATA",
            "time": mqtt_time(),
        }
        self._mqtt_client.publish(self._command_topic, json.dumps(payload))

    @abstractmethod
    def turn_on(self) -> None:
        """Turn on the device."""

    @abstractmethod
    def turn_off(self) -> None:
        """Turn off the device."""

    def set_speed(self, speed: int) -> None:
        """Set manual speed."""
        if not 1 <= speed <= 10:
            raise ValueError("Invalid speed %s", speed)
        self._set_speed(speed)

    @abstractmethod
    def _set_speed(self, speed: int) -> None:
        """Actually set the speed without range check."""

    @abstractmethod
    def enable_auto_mode(self) -> None:
        """Turn on auto mode."""

    @abstractmethod
    def disable_auto_mode(self) -> None:
        """Turn off auto mode."""

    @abstractmethod
    def enable_oscillation(self) -> None:
        """Turn on oscillation."""

    @abstractmethod
    def disable_oscillation(self) -> None:
        """Turn off oscillation."""

    def enable_night_mode(self) -> None:
        """Turn on auto mode."""
        self._set_configuration(nmod="ON")

    def disable_night_mode(self) -> None:
        """Turn off auto mode."""
        self._set_configuration(nmod="OFF")

    @abstractmethod
    def enable_continuous_monitoring(self) -> None:
        """Turn on continuous monitoring."""

    @abstractmethod
    def disable_continuous_monitoring(self) -> None:
        """Turn off continuous monitoring."""

    def set_sleep_timer(self, duration: int) -> None:
        """Set sleep timer."""
        if not 0 < duration <= 540:
            raise ValueError("Duration must be between 1 and 540")
        self._set_configuration(sltm="%04d" % duration)

    def disable_sleep_timer(self) -> None:
        """Disable sleep timer."""
        self._set_configuration(sltm="OFF")

    def reset_filter(self) -> None:
        """Reset filter life."""
        self._set_configuration(rstf="RSTF")


class DysonHeatingDevice(DysonFanDevice):
    """Dyson heating fan device."""

    @property
    def focus_mode(self) -> bool:
        """Return if fan focus mode is on."""
        return self._get_field_value(self._status, "ffoc") == "ON"

    @property
    def heat_target(self) -> float:
        """Return heat target in kelvin."""
        return int(self._get_field_value(self._status, "hmax")) / 10

    @property
    def heat_mode_is_on(self) -> bool:
        """Return if heat mode is set to on."""
        return self._get_field_value(self._status, "hmod") == "HEAT"

    @property
    def heat_status_is_on(self) -> bool:
        """Return if the device is currently heating."""
        return self._get_field_value(self._status, "hsta") == "HEAT"

    def set_heat_target(self, heat_target: float) -> None:
        """Set heat target in kelvin."""
        if not 274 <= heat_target <= 310:
            raise ValueError("Heat target must be between 274 and 310 kelvin")
        self._set_configuration(
            hmod="HEAT",
            hmax=f"{round(heat_target * 10):04d}",
        )

    def enable_heat_mode(self) -> None:
        """Enable heat mode."""
        self._set_configuration(hmod="HEAT")

    def disable_heat_mode(self) -> None:
        """Disable heat mode."""
        self._set_configuration(hmod="OFF")



"""Dyson 360 Eye cloud client."""

from datetime import datetime, timedelta
from enum import Enum
from typing import List, Optional

import attr

from .cloud_device import DysonCloudDevice


class CleaningType(Enum):
    """Cleaning type of the task."""

    Immediate = "Immediate"
    Manual = "Manual"
    Scheduled = "Scheduled"


@attr.s(auto_attribs=True, frozen=True)
class CleaningTask:
    """Represent a cleaning task."""

    cleaning_id: str
    start_time: datetime  # Local time without timezone info
    finish_time: datetime  # Local time without timezone info
    area: float  # In square meters
    charges: int
    cleaning_type: str
    is_interim: bool

    @classmethod
    def from_raw(cls, raw: dict):
        """Parse raw data from cloud API."""
        return cls(
            raw["Clean"],
            datetime.fromisoformat(raw["Started"]),
            datetime.fromisoformat(raw["Finished"]),
            raw["Area"],
            raw["Charges"],
            CleaningType(raw["Type"]),
            raw["IsInterim"],
        )

    @property
    def cleaning_time(self) -> timedelta:
        """Return the total cleaning time."""
        return self.finish_time - self.start_time


class DysonCloud360Eye(DysonCloudDevice):
    """Dyson 360 Eye cloud client."""

    def get_cleaning_history(self) -> List[CleaningTask]:
        """Get cleaning history from the cloud."""
        response = self._account.request(
            "GET",
            f"/v1/assets/devices/{self._serial}/cleanhistory",
        )
        return [CleaningTask.from_raw(raw) for raw in response.json()["Entries"]]

    def get_cleaning_map(self, cleaning_id: str) -> Optional[bytes]:
        """Get cleaning map in PNG format."""
        response = self._account.request(
            "GET",
            f"/v1/mapvisualizer/devices/{self._serial}/map/{cleaning_id}",
        )
        if response.status_code == 404:
            return None  # No map associate with the cleaning id
        return response.content



"""Dyson Pure Hot+Cool device."""

from .dyson_device import DysonHeatingDevice
from .dyson_pure_cool import DysonPureCool


class DysonPureHotCool(DysonPureCool, DysonHeatingDevice):
    """Dyson Pure Hot+Cool device."""



"""Test discovery."""
import socket
from typing import Optional
from unittest.mock import MagicMock, patch

import pytest
from zeroconf import ServiceBrowser, ServiceInfo, Zeroconf

from libdyson import DysonDiscovery
from libdyson.discovery import TYPE_DYSON_360_EYE, TYPE_DYSON_FAN
from libdyson.dyson_device import DysonDevice

from . import CREDENTIAL, HOST, SERIAL


@pytest.fixture(autouse=True)
def zeroconf_mock() -> MagicMock:
    """Return zeroconf mock."""
    zeroconf = MagicMock(spec=Zeroconf)
    with patch("libdyson.discovery.Zeroconf", return_value=zeroconf) as mocked:
        yield mocked


@pytest.fixture()
def service_browser_mock() -> MagicMock:
    """Return service browser mock."""
    service_browser = MagicMock(spec=ServiceBrowser)
    type(service_browser).zc = MagicMock()
    with patch(
        "libdyson.discovery.ServiceBrowser",
        return_value=service_browser,
    ) as mocked:
        yield mocked


@pytest.mark.parametrize("zeroconf_instance", [None, MagicMock(spec=Zeroconf)])
def test_discovery(
    service_browser_mock: MagicMock,
    zeroconf_mock: MagicMock,
    zeroconf_instance: Optional[Zeroconf],
):
    """Test discovery."""
    discovery = DysonDiscovery()
    device = DysonDevice(SERIAL, CREDENTIAL)
    discovery.start_discovery(zeroconf_instance)
    service_browser = service_browser_mock.return_value
    zeroconf = service_browser_mock.call_args.args[0]
    listener = service_browser_mock.call_args.args[2]

    if zeroconf_instance is None:
        zeroconf_mock.assert_called_once()
    else:
        zeroconf_mock.assert_not_called()

    # 360 Eye, discovered before registered
    name = f"360EYE-{SERIAL}.{TYPE_DYSON_360_EYE}"
    address = socket.inet_aton(HOST)
    service_info = ServiceInfo(TYPE_DYSON_360_EYE, name, addresses=[address])
    zeroconf.get_service_info = MagicMock(return_value=service_info)
    listener.add_service(zeroconf, TYPE_DYSON_360_EYE, name)
    callback = MagicMock()
    discovery.register_device(device, callback)
    callback.assert_called_once_with(HOST)

    # Pure Cool Link, registered before discovered
    serial = "NK6-CN-HAA0000A"
    device = DysonDevice(serial, CREDENTIAL)
    callback = MagicMock()
    discovery.register_device(device, callback)
    name = f"475_{serial}.{TYPE_DYSON_FAN}"
    service_info = ServiceInfo(TYPE_DYSON_FAN, name, addresses=[address])
    zeroconf.get_service_info = MagicMock(return_value=service_info)
    listener.add_service(zeroconf, TYPE_DYSON_FAN, name)
    callback.assert_called_once_with(HOST)

    # Stop discovery
    discovery.stop_discovery()
    service_browser.cancel.assert_called_once()
    service_browser.zc.close.assert_called_once()


def test_stop_runtime_error():
    """Test ServiceBrowser.cancel RuntimeError handling."""
    service_browser = MagicMock(spec=ServiceBrowser)
    type(service_browser).zc = MagicMock()
    type(service_browser).cancel = MagicMock(side_effect=RuntimeError)
    with patch(
        "libdyson.discovery.ServiceBrowser",
        return_value=service_browser,
    ):
        discovery = DysonDiscovery()
        discovery.start_discovery()
        discovery.stop_discovery()
        service_browser.cancel.assert_called_once()
        service_browser.zc.close.assert_called_once()



"""Account region list."""

REGIONS = {
    "AU": "Australia",
    "AT": "Austria",
    "BE": "Belgium",
    "CA": "Canada",
    "HR": "Croatia",
    "CZ": "Czechia",
    "DK": "Denmark",
    "FI": "Finland",
    "FR": "France",
    "DE": "Germany",
    "HK": "Hong Kong",
    "HU": "Hungary",
    "IN": "India",
    "ID": "Indonesia",
    "IE": "Ireland",
    "IL": "Israel",
    "IT": "Italy",
    "JP": "Japan",
    "LT": "Lithuania",
    "CN": "Mainland China",
    "MY": "Malaysia",
    "MX": "Mexico",
    "NL": "Netherlands",
    "NZ": "New Zealand",
    "NO": "Norway",
    "PH": "Philippines",
    "PL": "Poland",
    "PT": "Portugal",
    "RO": "Romania",
    "SA": "Saudi Arabia",
    "SG": "Singapore",
    "SI": "Slovenia",
    "KR": "South Korea",
    "ES": "Spain",
    "SE": "Sweden",
    "CH": "Switzerland",
    "TW": "Taiwan",
    "TH": "Thailand",
    "TR": "Turkey",
    "AE": "United Arab Emirates",
    "GB": "United Kingdom",
    "US": "United States of America",
}



"""Tests for Dyson Pure Cool Link fan."""

import pytest

from libdyson import DEVICE_TYPE_PURE_COOL_LINK_DESK
from libdyson.const import ENVIRONMENTAL_INIT, AirQualityTarget
from libdyson.dyson_pure_cool_link import DysonPureCoolLink

from . import CREDENTIAL, HOST, SERIAL
from .mocked_mqtt import MockedMQTT
from .test_fan_device import assert_command

DEVICE_TYPE = DEVICE_TYPE_PURE_COOL_LINK_DESK

STATUS = {
    "mode-reason": "RAPP",
    "state-reason": "MODE",
    "dial": "OFF",
    "rssi": "-29",
    "product-state": {
        "fmod": "OFF",
        "fnst": "OFF",
        "fnsp": "0001",
        "qtar": "0002",
        "oson": "OFF",
        "rhtm": "OFF",
        "filf": "1500",
        "ercd": "NONE",
        "nmod": "OFF",
        "wacd": "NONE",
    },
    "scheduler": {"srsc": "8773", "dstv": "0000", "tzid": "0001"},
}

ENVIRONMENTAL_DATA = {
    "data": {
        "tact": "OFF",
        "hact": "OFF",
        "pact": "0003",
        "vact": "0004",
        "sltm": "OFF",
    }
}


def test_properties(mqtt_client: MockedMQTT):
    """Test properties of Pure Cool Link."""
    device = DysonPureCoolLink(SERIAL, CREDENTIAL, DEVICE_TYPE)
    device.connect(HOST)

    # Status
    assert device.is_on is False
    assert device.auto_mode is False
    assert device.oscillation is False
    assert device.air_quality_target == AirQualityTarget.DEFAULT
    assert device.filter_life == 1500

    # Environmental
    assert device.particulates == 3
    assert device.volatile_organic_compounds == 4

    new_status = {
        "mode-reason": "LAPP",
        "state-reason": "MODE",
        "product-state": {
            "fmod": ["OFF", "AUTO"],
            "fnst": ["OFF", "FAN"],
            "fnsp": ["0001", "AUTO"],
            "qtar": ["0002", "0001"],
            "oson": ["OFF", "ON"],
            "rhtm": ["OFF", "ON"],
            "filf": ["1500", "1450"],
            "ercd": ["NONE", "NONE"],
            "nmod": ["OFF", "ON"],
            "wacd": ["NONE", "NONE"],
        },
        "scheduler": {"srsc": "8773", "dstv": "0000", "tzid": "0001"},
    }
    mqtt_client.state_change(new_status)
    assert device.is_on is True
    assert device.auto_mode is True
    assert device.oscillation is True
    assert device.air_quality_target == AirQualityTarget.VERY_SENSITIVE
    assert device.filter_life == 1450

    mqtt_client._environmental_data = {
        "data": {
            "tact": "2903",
            "hact": "0030",
            "pact": "0005",
            "vact": "INIT",
            "sltm": "0003",
        }
    }
    device.request_environmental_data()
    assert device.particulates == 5
    assert device.volatile_organic_compounds == ENVIRONMENTAL_INIT


@pytest.mark.parametrize(
    "command,command_args,msg_data",
    [
        ("turn_on", [], {"fmod": "FAN"}),
        ("turn_off", [], {"fmod": "OFF"}),
        ("set_speed", [3], {"fmod": "FAN", "fnsp": "0003"}),
        ("enable_auto_mode", [], {"fmod": "AUTO"}),
        ("disable_auto_mode", [], {"fmod": "FAN"}),
        ("enable_oscillation", [], {"oson": "ON"}),
        ("disable_oscillation", [], {"oson": "OFF"}),
        ("enable_continuous_monitoring", [], {"fmod": "OFF", "rhtm": "ON"}),
        ("disable_continuous_monitoring", [], {"fmod": "OFF", "rhtm": "OFF"}),
        ("set_air_quality_target", [AirQualityTarget.SENSITIVE], {"qtar": "0003"}),
    ],
)
def test_command(
    mqtt_client: MockedMQTT,
    command: str,
    command_args: list,
    msg_data: dict,
):
    """Test commands of Pure Cool Link."""
    assert_command(
        DysonPureCoolLink(SERIAL, CREDENTIAL, DEVICE_TYPE),
        mqtt_client,
        command,
        command_args,
        msg_data,
    )



"""Test DysonDevice functionalities."""
from unittest.mock import MagicMock, patch

import pytest

from libdyson.const import MessageType
from libdyson.dyson_device import DysonDevice
from libdyson.exceptions import (
    DysonConnectionRefused,
    DysonConnectTimeout,
    DysonInvalidCredential,
    DysonNotConnected,
)

from . import CREDENTIAL, HOST, SERIAL
from .mocked_mqtt import MockedMQTT

DEVICE_TYPE = "device_type"

STATUS = {
    "key1": "V1",
    "key2": "V2",
}


class _TestDevice(DysonDevice):
    def __init__(self, serial: str, credential: str):
        """Initialize the device."""
        super().__init__(serial, credential)

    @property
    def device_type(self) -> str:
        return DEVICE_TYPE

    @property
    def _status_topic(self) -> str:
        return f"{DEVICE_TYPE}/{self._serial}/status"

    def _update_status(self, payload: dict) -> None:
        payload.pop("msg")
        payload.pop("time")
        self._status = payload


@pytest.fixture(autouse=True)
def mqtt_client() -> MockedMQTT:
    """Return mocked mqtt client."""
    mocked_mqtt = MockedMQTT(
        HOST,
        SERIAL,
        CREDENTIAL,
        f"{DEVICE_TYPE}/{SERIAL}/command",
        f"{DEVICE_TYPE}/{SERIAL}/status",
        STATUS,
    )
    with patch("libdyson.dyson_device.mqtt.Client", mocked_mqtt.refersh), patch(
        "libdyson.dyson_device.TIMEOUT", 0
    ):
        yield mocked_mqtt


def test_properties():
    """Test device properties."""
    device = _TestDevice(SERIAL, CREDENTIAL)
    assert device.serial == SERIAL
    assert device.is_connected is False
    assert device.device_type == DEVICE_TYPE


def test_connect(mqtt_client: MockedMQTT):
    """Test successful connection."""
    device = _TestDevice(SERIAL, CREDENTIAL)
    device.connect(HOST)
    assert device.is_connected is True
    assert mqtt_client.connected is True
    assert mqtt_client.loop_started is True


def test_invalid_credential(mqtt_client: MockedMQTT):
    """Test invalid credential."""
    device = _TestDevice(SERIAL, "invalid")
    with pytest.raises(DysonInvalidCredential):
        device.connect(HOST)
    assert device.is_connected is False
    assert mqtt_client.loop_started is False


def test_connect_timeout(mqtt_client: MockedMQTT):
    """Test connection timed out."""
    device = _TestDevice(SERIAL, CREDENTIAL)
    with pytest.raises(DysonConnectTimeout):
        device.connect("192.168.1.5")
    assert device.is_connected is False
    assert mqtt_client.loop_started is False


def test_connect_status_timeout(mqtt_client: MockedMQTT):
    """Test first data timed out."""

    def _publish(topic: str, payload: str) -> None:
        pass  # Do nothing in publish so CURRENT-STATUS never sent

    mqtt_client.publish = _publish
    device = _TestDevice(SERIAL, CREDENTIAL)
    with pytest.raises(DysonConnectTimeout):
        device.connect(HOST)
    assert device.is_connected is False
    assert mqtt_client.connected is False
    assert mqtt_client.loop_started is False


def test_connection_refused(mqtt_client: MockedMQTT):
    """Test connection refused."""

    def _connect_async(host: str) -> None:
        mqtt_client.on_connect(mqtt_client, None, None, 2)
        mqtt_client.on_disconnect(mqtt_client, None, 3)

    mqtt_client.connect_async = _connect_async
    device = _TestDevice(SERIAL, CREDENTIAL)
    with pytest.raises(DysonConnectionRefused):
        device.connect(HOST)
    assert device.is_connected is False
    assert mqtt_client.connected is False
    assert mqtt_client.loop_started is False


def test_disconnect(mqtt_client: MockedMQTT):
    """Test successful disconnect."""
    device = _TestDevice(SERIAL, CREDENTIAL)
    device.connect(HOST)
    device.disconnect()
    assert device.is_connected is False
    assert mqtt_client.connected is False
    assert mqtt_client.loop_started is False


def test_disconnect_timeout(mqtt_client: MockedMQTT):
    """Test disconnection timed out."""

    def _disconnect():
        pass  # Do nothing so on_disconnect never called

    mqtt_client.disconnect = _disconnect
    device = _TestDevice(SERIAL, CREDENTIAL)
    device.connect(HOST)
    device.disconnect()
    assert device.is_connected is False
    assert mqtt_client.loop_started is False


def test_not_connected():
    """Test send commands without connection."""
    device = _TestDevice(SERIAL, CREDENTIAL)
    with pytest.raises(DysonNotConnected):
        device.request_current_status()
    assert device._status is None
    with pytest.raises(DysonNotConnected):
        device._send_command("COMMAND")


def test_status_update(mqtt_client: MockedMQTT):
    """Test status update."""
    device = _TestDevice(SERIAL, CREDENTIAL)
    callback = MagicMock()
    device.add_message_listener(callback)
    device.connect(HOST)

    # Data updated
    assert device._status == STATUS
    callback.assert_called_once_with(MessageType.STATE)
    callback.reset_mock()

    new_status = {
        "key1": "V2",
        "key2": "V1",
    }
    mqtt_client.state_change(new_status)
    assert device._status == new_status
    callback.assert_called_once_with(MessageType.STATE)
    callback.reset_mock()

    # Remove callback
    device.remove_message_listener(callback)
    mqtt_client.state_change(new_status)
    callback.assert_not_called()



"""Tests for Pure Cool Formaldehyde"""

import pytest

from libdyson.const import (
    DEVICE_TYPE_PURE_COOL_FORMALDEHYDE,
    ENVIRONMENTAL_FAIL,
    ENVIRONMENTAL_INIT,
    ENVIRONMENTAL_OFF,
)
from libdyson.dyson_pure_cool import DysonPureCoolFormaldehyde

from . import CREDENTIAL, HOST, SERIAL
from .mocked_mqtt import MockedMQTT
from .test_pure_cool import STATUS as TEST_PURE_COOL_STATUS

DEVICE_TYPE = DEVICE_TYPE_PURE_COOL_FORMALDEHYDE

STATUS = TEST_PURE_COOL_STATUS
ENVIRONMENTAL_DATA = {
    "data": {
        "tact": "OFF",
        "hact": "OFF",
        "pm25": "OFF",
        "pm10": "OFF",
        "va10": "INIT",
        "noxl": "FAIL",
        "p25r": "OFF",
        "p10r": "OFF",
        "sltm": "OFF",
        "hcho": "OFF",
        "hchr": "OFF",
    }
}


def test_properties(mqtt_client: MockedMQTT):
    """Test properties of Pure Cool Link Formaldehyde."""
    device = DysonPureCoolFormaldehyde(SERIAL, CREDENTIAL, DEVICE_TYPE)
    device.connect(HOST)

    # Environmental
    assert device.particulate_matter_2_5 == ENVIRONMENTAL_OFF
    assert device.particulate_matter_10 == ENVIRONMENTAL_OFF
    assert device.volatile_organic_compounds == ENVIRONMENTAL_INIT
    assert device.nitrogen_dioxide == ENVIRONMENTAL_FAIL
    assert device.formaldehyde == ENVIRONMENTAL_OFF

    mqtt_client._environmental_data = {
        "data": {
            "tact": "2977",
            "hact": "0058",
            "pm25": "0009",
            "pm10": "0005",
            "va10": "0004",
            "noxl": "0011",
            "p25r": "0010",
            "p10r": "0009",
            "sltm": "OFF",
            "hcho": "0001",
            "hchr": "0002",
        }
    }
    device.request_environmental_data()
    assert device.particulate_matter_2_5 == 9
    assert device.particulate_matter_10 == 5
    assert device.volatile_organic_compounds == 4
    assert device.nitrogen_dioxide == 11
    assert device.formaldehyde == 1



"""Dyson device discovery."""

import socket
import threading
from typing import Callable, Optional

from zeroconf import ServiceBrowser, ServiceInfo, Zeroconf

from .dyson_device import DysonDevice

TYPE_DYSON_360_EYE = "_360eye_mqtt._tcp.local."
TYPE_DYSON_FAN = "_dyson_mqtt._tcp.local."


class DysonDiscovery:
    """Dyson device discovery."""

    def __init__(self):
        """Initialize the instance."""
        self._registered = {}
        self._discovered = {}
        self._lock = threading.Lock()
        self._browser = None

    def register_device(
        self, device: DysonDevice, callback: Callable[[str], None]
    ) -> None:
        """Register a device."""
        with self._lock:
            if device.serial in self._discovered:
                callback(self._discovered[device.serial])
            else:
                self._registered[device.serial] = callback

    def device_discovered(self, info: ServiceInfo) -> None:
        """Call when a device is discovered."""
        if info.type == TYPE_DYSON_360_EYE:
            serial = (info.name.split(".")[0]).split("-", 1)[1]
        else:  # TYPE_DYSON_FAN
            serial = (info.name.split(".")[0]).split("_")[1]
        address = socket.inet_ntoa(info.addresses[0])
        with self._lock:
            if serial in self._registered:
                callback = self._registered.pop(serial)
                callback(address)
            else:
                self._discovered[serial] = address

    def start_discovery(self, zeroconf_instance: Optional[Zeroconf] = None) -> None:
        """Start discovery."""
        listener = DysonListener(self)
        zeroconf = zeroconf_instance or Zeroconf()
        self._browser = ServiceBrowser(
            zeroconf,
            [TYPE_DYSON_360_EYE, TYPE_DYSON_FAN],
            listener,
        )

    def stop_discovery(self) -> None:
        """Stop discovery."""
        try:
            self._browser.cancel()
        except RuntimeError:
            # Throws when called from callback
            # cannot join current thread
            pass
        self._browser.zc.close()
        self._browser = None


class DysonListener:
    """Listener for zeroconf events."""

    def __init__(self, dyson_discovery: DysonDiscovery):
        """Initialize the listener."""
        self._dyson_discovery = dyson_discovery

    def add_service(self, zeroconf: Zeroconf, type: str, name: str) -> None:
        """Add a new service."""
        info = zeroconf.get_service_info(type, name)
        self._dyson_discovery.device_discovered(info)

    def update_service(self, zeroconf: Zeroconf, type: str, name: str) -> None:
        """Update a service."""
        # Currently not doing anything

    def remove_service(self, zeroconf: Zeroconf, type: str, name: str) -> None:
        """Remove a service."""
        # Currently not doing anything



"""Setup script."""
import setuptools

REQUIRES = [
    "paho_mqtt",
    "cryptography>=3.1",
    "requests",
    "zeroconf",
    "attrs",
]

setuptools.setup(
    include_package_data=True,
    install_requires=REQUIRES,
)



"""Dyson cloud client utilities."""

import base64
import json

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

DYSON_ENCRYPTION_KEY = (
    b"\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10"
    b"\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f "
)
DYSON_ENCRYPTION_INIT_VECTOR = (
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
)


def _unpad(string: str) -> str:
    """Un-pad string."""
    return string[: -ord(string[len(string) - 1 :])]


def decrypt_password(encrypted_password: str) -> str:
    """Decrypt local credential into MQTT password."""
    cipher = Cipher(
        algorithms.AES(DYSON_ENCRYPTION_KEY),
        modes.CBC(DYSON_ENCRYPTION_INIT_VECTOR),
    )
    decryptor = cipher.decryptor()
    encrypted = base64.b64decode(encrypted_password)
    decrypted = decryptor.update(encrypted) + decryptor.finalize()
    json_password = json.loads(_unpad(decrypted))
    return json_password["apPasswordHash"]



"""Tests for Dyson 360 Heurist vacuum."""
from unittest.mock import patch

import pytest

from libdyson import (
    DEVICE_TYPE_360_HEURIST,
    CleaningMode,
    Dyson360Heurist,
    VacuumHeuristPowerMode,
)

from . import CREDENTIAL, HOST, SERIAL
from .mocked_mqtt import MockedMQTT

STATUS = {
    "currentVacuumPowerMode": "1",
    "defaultVacuumPowerMode": "2",
    "currentCleaningMode": "zoneConfigured",
    "defaultCleaningMode": "global",
}


@pytest.fixture(autouse=True)
def mqtt_client() -> MockedMQTT:
    """Return mocked mqtt client."""
    mocked_mqtt = MockedMQTT(
        HOST,
        SERIAL,
        CREDENTIAL,
        f"{DEVICE_TYPE_360_HEURIST}/{SERIAL}/command",
        f"{DEVICE_TYPE_360_HEURIST}/{SERIAL}/status",
        STATUS,
    )
    with patch("libdyson.dyson_device.mqtt.Client", mocked_mqtt.refersh):
        yield mocked_mqtt


def test_properties(mqtt_client: MockedMQTT):
    """Test properties of 360 Heurist."""
    device = Dyson360Heurist(SERIAL, CREDENTIAL)
    device.connect(HOST)
    assert device.current_power_mode == VacuumHeuristPowerMode.QUIET
    assert device.default_power_mode == VacuumHeuristPowerMode.HIGH
    assert device.current_cleaning_mode == CleaningMode.ZONE_CONFIGURED
    assert device.default_cleaning_mode == CleaningMode.GLOBAL
    assert device.is_bin_full is False

    new_status = {
        "currentVacuumPowerMode": "2",
        "defaultVacuumPowerMode": "3",
        "currentCleaningMode": "global",
        "defaultCleaningMode": "zoneConfigured",
        "faults": {
            "AIRWAYS": {"active": True, "description": "1.0.-1"},
        },
    }
    mqtt_client.state_change(new_status)
    assert device.current_power_mode == VacuumHeuristPowerMode.HIGH
    assert device.default_power_mode == VacuumHeuristPowerMode.MAX
    assert device.current_cleaning_mode == CleaningMode.GLOBAL
    assert device.default_cleaning_mode == CleaningMode.ZONE_CONFIGURED
    assert device.is_bin_full is True


@pytest.mark.parametrize(
    "command,command_args,msg,msg_data",
    [
        (
            "start_all_zones",
            [],
            "START",
            {"cleaningMode": "global", "fullCleanType": "immediate"},
        ),
        ("pause", [], "PAUSE", {}),
        ("resume", [], "RESUME", {}),
        ("abort", [], "ABORT", {}),
        (
            "set_default_power_mode",
            [VacuumHeuristPowerMode.QUIET],
            "STATE-SET",
            {"defaults": {"defaultVacuumPowerMode": "1"}},
        ),
        (
            "set_default_power_mode",
            [VacuumHeuristPowerMode.HIGH],
            "STATE-SET",
            {"defaults": {"defaultVacuumPowerMode": "2"}},
        ),
        (
            "set_default_power_mode",
            [VacuumHeuristPowerMode.MAX],
            "STATE-SET",
            {"defaults": {"defaultVacuumPowerMode": "3"}},
        ),
    ],
)
def test_command(
    mqtt_client: MockedMQTT, command: str, command_args: list, msg: str, msg_data: dict
):
    """Test commands of 360 Heurist."""
    device = Dyson360Heurist(SERIAL, CREDENTIAL)
    device.connect(HOST)

    func = getattr(device, command)
    func(*command_args)
    len(mqtt_client.commands) == 1
    payload = mqtt_client.commands[0]
    assert payload.pop("msg") == msg
    assert payload.pop("mode-reason") == "LAPP"
    payload.pop("time")
    assert payload == msg_data



"""Dyson cloud test configuration."""

from unittest.mock import patch

import pytest

from .mocked_requests import MockedRequests


@pytest.fixture()
def mocked_requests() -> MockedRequests:
    """Return mocked requests library."""
    mocked_requests = MockedRequests()

    with patch("libdyson.cloud.account.requests.request", mocked_requests.request):
        yield mocked_requests



"""Tests for DysonFanDevice."""

import json
from typing import Optional
from unittest.mock import MagicMock

import pytest

from libdyson.const import (
    DEVICE_TYPE_PURE_COOL_LINK_DESK,
    ENVIRONMENTAL_OFF,
    MessageType,
)
from libdyson.dyson_device import DysonFanDevice
from libdyson.exceptions import DysonConnectTimeout, DysonNotConnected

from . import CREDENTIAL, HOST, SERIAL
from .mocked_mqtt import MockedMQTT

DEVICE_TYPE = DEVICE_TYPE_PURE_COOL_LINK_DESK

STATUS = {
    "mode-reason": "RAPP",
    "state-reason": "MODE",
    "dial": "OFF",
    "rssi": "-29",
    "product-state": {
        "fnst": "OFF",
        "fnsp": "0001",
        "rhtm": "OFF",
        "ercd": "NONE",
        "nmod": "OFF",
        "wacd": "NONE",
    },
    "scheduler": {"srsc": "8773", "dstv": "0000", "tzid": "0001"},
}

ENVIRONMENTAL_DATA = {
    "data": {
        "tact": "OFF",
        "hact": "OFF",
        "sltm": "OFF",
    }
}


def test_properties(mqtt_client: MockedMQTT):
    """Test properties of fan device."""
    device = DysonFanDevice(SERIAL, CREDENTIAL, DEVICE_TYPE)
    device.connect(HOST)

    # Status
    assert device.device_type == DEVICE_TYPE
    assert device.fan_state is False
    assert device.speed == 1
    assert device.night_mode is False
    assert device.continuous_monitoring is False
    assert device.error_code == "NONE"
    assert device.warning_code == "NONE"

    # Environmental
    assert device.humidity == ENVIRONMENTAL_OFF
    assert device.temperature == ENVIRONMENTAL_OFF
    assert device.sleep_timer == ENVIRONMENTAL_OFF

    error_code = "0X03"  # Just mock data
    warning_code = "0X01"  # Just mock data
    new_status = {
        "mode-reason": "LAPP",
        "state-reason": "MODE",
        "product-state": {
            "fnst": ["OFF", "FAN"],
            "fnsp": ["0001", "AUTO"],
            "rhtm": ["OFF", "ON"],
            "ercd": ["NONE", error_code],
            "nmod": ["OFF", "ON"],
            "wacd": ["NONE", warning_code],
        },
        "scheduler": {"srsc": "8773", "dstv": "0000", "tzid": "0001"},
    }
    mqtt_client.state_change(new_status)
    assert device.fan_state is True
    assert device.speed is None
    assert device.night_mode is True
    assert device.continuous_monitoring is True
    assert device.error_code == error_code
    assert device.warning_code == warning_code

    mqtt_client._environmental_data = {
        "data": {
            "tact": "2903",
            "hact": "0030",
            "sltm": "0003",
        }
    }
    device.request_environmental_data()
    assert device.humidity == 30
    assert device.temperature == 290.3
    assert device.sleep_timer == 3


def assert_command(
    device: DysonFanDevice,
    mqtt_client: MockedMQTT,
    command: str,
    command_args: list,
    msg_data: dict,
    new_status: Optional[dict] = None,
):
    """Test commands of fan device."""
    device.connect(HOST)
    if new_status is not None:
        mqtt_client.state_change(new_status)
    func = getattr(device, command)
    func(*command_args)
    assert len(mqtt_client.commands) == 1
    payload = mqtt_client.commands[0]
    assert payload["msg"] == "STATE-SET"
    assert payload["data"] == msg_data


@pytest.mark.parametrize(
    "command,command_args,msg_data",
    [
        ("enable_night_mode", [], {"nmod": "ON"}),
        ("disable_night_mode", [], {"nmod": "OFF"}),
        ("set_sleep_timer", [15], {"sltm": "0015"}),
        ("disable_sleep_timer", [], {"sltm": "OFF"}),
        ("reset_filter", [], {"rstf": "RSTF"}),
    ],
)
def test_command(
    mqtt_client: MockedMQTT,
    command: str,
    command_args: list,
    msg_data: dict,
):
    """Test commands of fan device."""
    assert_command(
        DysonFanDevice(SERIAL, CREDENTIAL, DEVICE_TYPE),
        mqtt_client,
        command,
        command_args,
        msg_data,
    )


@pytest.mark.parametrize(
    "command,command_args",
    [
        ("set_speed", [0]),
        ("set_speed", [11]),
        ("set_sleep_timer", [0]),
        ("set_sleep_timer", [600]),
    ],
)
def test_command_invalid_data(
    mqtt_client: MockedMQTT,
    command: str,
    command_args: list,
):
    """Test commands with invalid data."""
    device = DysonFanDevice(SERIAL, CREDENTIAL, DEVICE_TYPE)
    device.connect(HOST)
    func = getattr(device, command)
    with pytest.raises(ValueError):
        func(*command_args)
    assert len(mqtt_client.commands) == 0


def test_not_connected(mqtt_client: MockedMQTT):
    """Test making operations without connection."""
    device = DysonFanDevice(SERIAL, CREDENTIAL, DEVICE_TYPE)
    with pytest.raises(DysonNotConnected):
        device.enable_night_mode()
    with pytest.raises(DysonNotConnected):
        device.request_environmental_data()


def test_connect_environmental_timeout(mqtt_client: MockedMQTT):
    """Test environmental data timed out during connection."""
    original_publish = mqtt_client.publish

    def _publish(topic: str, payload: str) -> None:
        if (
            json.loads(payload)["msg"]
            == "REQUEST-PRODUCT-ENVIRONMENT-CURRENT-SENSOR-DATA"
        ):
            return  # Do not publish environmental data request
        original_publish(topic, payload)

    mqtt_client.publish = _publish

    device = DysonFanDevice(SERIAL, CREDENTIAL, DEVICE_TYPE)
    with pytest.raises(DysonConnectTimeout):
        device.connect(HOST)
    assert device._status_data_available.is_set()
    assert device.is_connected is False
    assert mqtt_client.connected is False
    assert mqtt_client.loop_started is False


def test_environmental_callback(mqtt_client: MockedMQTT):
    """Test callback on environmental data."""
    device = DysonFanDevice(SERIAL, CREDENTIAL, DEVICE_TYPE)
    callback = MagicMock()
    device.add_message_listener(callback)
    device.connect(HOST)
    callback.assert_called_with(MessageType.ENVIRONMENTAL)
    callback.reset_mock()

    device.request_environmental_data()
    callback.assert_called_once_with(MessageType.ENVIRONMENTAL)
    callback.reset_mock()
